#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from heat.common import exception
from heat.engine import attributes
from heat.engine import status


class StackDefinition(object):
    """Class representing the definition of a Stack, but not its current state.

    This is the interface through which template functions will access data
    about the stack definition, including the template and current values of
    resource reference IDs and attributes.

    This API can be considered stable by third-party Template or Function
    plugins, and no part of it should be changed or removed without an
    appropriate deprecation process.
    """

    def __init__(self, context, template, stack_identifier, resource_data,
                 parent_info=None):
        self._context = context
        self._template = template
        self._resource_data = {} if resource_data is None else resource_data
        self._parent_info = parent_info
        self._zones = None
        self.parameters = template.parameters(stack_identifier,
                                              template.env.params,
                                              template.env.param_defaults)
        self._resource_defns = None
        self._resources = {}
        self._output_defns = None

    def clone_with_new_template(self, new_template, stack_identifier,
                                clear_resource_data=False):
        """Create a new StackDefinition with a different template."""
        res_data = {} if clear_resource_data else dict(self._resource_data)
        return type(self)(self._context, new_template, stack_identifier,
                          res_data, self._parent_info)

    @property
    def t(self):
        """The stack's template."""
        return self._template

    @property
    def env(self):
        """The stack's environment."""
        return self._template.env

    def _load_rsrc_defns(self):
        self._resource_defns = self._template.resource_definitions(self)

    def resource_definition(self, resource_name):
        """Return the definition of the given resource."""
        if self._resource_defns is None:
            self._load_rsrc_defns()
        return self._resource_defns[resource_name]

    def enabled_rsrc_names(self):
        """Return the set of names of all enabled resources in the template."""
        if self._resource_defns is None:
            self._load_rsrc_defns()
        return set(self._resource_defns)

    def _load_output_defns(self):
        self._output_defns = self._template.outputs(self)

    def output_definition(self, output_name):
        """Return the definition of the given output."""
        if self._output_defns is None:
            self._load_output_defns()
        return self._output_defns[output_name]

    def enabled_output_names(self):
        """Return the set of names of all enabled outputs in the template."""
        if self._output_defns is None:
            self._load_output_defns()
        return set(self._output_defns)

    def all_rsrc_names(self):
        """Return the set of names of all resources in the template.

        This includes resources that are disabled due to false conditionals.
        """
        if hasattr(self._template, 'RESOURCES'):
            return set(self._template.get(self._template.RESOURCES,
                                          self._resource_defns or []))
        else:
            return self.enabled_rsrc_names()

    def all_resource_types(self):
        """Return the set of types of all resources in the template."""
        if self._resource_defns is None:
            self._load_rsrc_defns()
        return set(self._resource_defns[res].resource_type
                   for res in self._resource_defns)

    def get_availability_zones(self):
        """Return the list of Nova availability zones."""
        if self._zones is None:
            nova = self._context.clients.client('nova')
            zones = nova.availability_zones.list(detailed=False)
            self._zones = [zone.zoneName for zone in zones]
        return self._zones

    def __contains__(self, resource_name):
        """Return True if the given resource name is present and enabled."""
        if self._resource_defns is not None:
            return resource_name in self._resource_defns
        else:
            # In Cfn templates, we need to know whether Ref refers to a
            # resource or a parameter in order to parse the resource
            # definitions
            return resource_name in self._template[self._template.RESOURCES]

    def __getitem__(self, resource_name):
        """Return a proxy for the given resource."""
        if resource_name not in self._resources:
            res_proxy = ResourceProxy(resource_name,
                                      self.resource_definition(resource_name),
                                      self._resource_data.get(resource_name))
            self._resources[resource_name] = res_proxy
        return self._resources[resource_name]

    @property
    def parent_resource(self):
        """Return a proxy for the parent resource.

        Returns None if the stack is not a provider stack for a
        TemplateResource.
        """
        return self._parent_info


class ResourceProxy(status.ResourceStatus):
    """A lightweight API for essential data about a resource.

    This is the interface through which template functions will access data
    about particular resources in the stack definition, such as the resource
    definition and current values of reference IDs and attributes.

    Resource proxies for some or all resources in the stack will potentially be
    loaded for every check resource operation, so it is essential that this API
    is implemented efficiently, using only the data received over RPC and
    without reference to the resource data stored in the database.

    This API can be considered stable by third-party Template or Function
    plugins, and no part of it should be changed or removed without an
    appropriate deprecation process.
    """

    __slots__ = ('name', '_definition', '_resource_data')

    def __init__(self, name, definition, resource_data):
        self.name = name
        self._definition = definition
        self._resource_data = resource_data

    @property
    def t(self):
        """The resource definition."""
        return self._definition

    def _res_data(self):
        assert self._resource_data is not None, "Resource data not available"
        return self._resource_data

    @property
    def attributes_schema(self):
        """A set of the valid top-level attribute names.

        This is provided for backwards-compatibility for functions that require
        a container with all of the valid attribute names in order to validate
        the template. Other operations on it are invalid because we don't
        actually have access to the attributes schema here; hence we return a
        set instead of a dict.
        """
        return set(self._res_data().attribute_names())

    @property
    def external_id(self):
        """The external ID of the resource."""
        return self._definition.external_id()

    @property
    def state(self):
        """The current state (action, status) of the resource."""
        return self.action, self.status

    @property
    def action(self):
        """The current action of the resource."""
        if self._resource_data is None:
            return self.INIT
        return self._resource_data.action

    @property
    def status(self):
        """The current status of the resource."""
        if self._resource_data is None:
            return self.COMPLETE
        return self._resource_data.status

    def FnGetRefId(self):
        """For the intrinsic function get_resource."""
        if self._resource_data is None:
            return self.name
        return self._resource_data.reference_id()

    def FnGetAtt(self, attr, *path):
        """For the intrinsic function get_attr."""
        if path:
            attr = (attr,) + path
        try:
            return self._res_data().attribute(attr)
        except KeyError:
            raise exception.InvalidTemplateAttribute(resource=self.name,
                                                     key=attr)

    def FnGetAtts(self):
        """For the intrinsic function get_attr when getting all attributes.

        :returns: a dict of all of the resource's attribute values, excluding
                  the "show" attribute.
        """
        all_attrs = self._res_data().attributes()
        return dict((k, v) for k, v in all_attrs.items()
                    if k != attributes.SHOW_ATTR)


def update_resource_data(stack_definition, resource_name, resource_data):
    """Store new resource state data for the specified resource.

    This function enables the legacy (non-convergence) path to store updated
    NodeData as resources are created/updated in a single StackDefinition
    that lasts for the entire lifetime of the stack operation.
    """
    stack_definition._resource_data[resource_name] = resource_data
    stack_definition._resources.pop(resource_name, None)

    # Clear the cached dep_attrs for any resource or output that directly
    # depends on the resource whose data we are updating. This ensures that if
    # any of the data we just updated is referenced in the path of a get_attr
    # function, future calls to dep_attrs() will reflect this new data.
    res_defns = stack_definition._resource_defns or {}
    op_defns = stack_definition._output_defns or {}

    all_defns = itertools.chain(res_defns.values(),
                                op_defns.values())
    for defn in all_defns:
        if resource_name in defn.required_resource_names():
            defn._all_dep_attrs = None


def add_resource(stack_definition, resource_definition):
    """Insert the given resource definition into the stack definition.

    Add the resource to the template and store any temporary data.
    """
    resource_name = resource_definition.name
    stack_definition._resources.pop(resource_name, None)
    stack_definition._resource_data.pop(resource_name, None)
    stack_definition.t.add_resource(resource_definition)
    if stack_definition._resource_defns is not None:
        stack_definition._resource_defns[resource_name] = resource_definition


def remove_resource(stack_definition, resource_name):
    """Remove the named resource from the stack definition.

    Remove the resource from the template and eliminate references to it.
    """
    stack_definition.t.remove_resource(resource_name)
    if stack_definition._resource_defns is not None:
        stack_definition._resource_defns.pop(resource_name, None)
    stack_definition._resource_data.pop(resource_name, None)
    stack_definition._resources.pop(resource_name, None)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from stevedore import extension

from heat.common import pluginutils
from heat.engine import clients
from heat.engine import environment
from heat.engine import plugin_manager


def _register_resources(env, type_pairs):
    for res_name, res_class in type_pairs:
        env.register_class(res_name, res_class)


def _register_constraints(env, type_pairs):
    for constraint_name, constraint in type_pairs:
        env.register_constraint(constraint_name, constraint)


def _register_stack_lifecycle_plugins(env, type_pairs):
    for stack_lifecycle_name, stack_lifecycle_class in type_pairs:
        env.register_stack_lifecycle_plugin(stack_lifecycle_name,
                                            stack_lifecycle_class)


def _register_event_sinks(env, type_pairs):
    for sink_name, sink_class in type_pairs:
        env.register_event_sink(sink_name, sink_class)


def _get_mapping(namespace):
    mgr = extension.ExtensionManager(
        namespace=namespace,
        invoke_on_load=False,
        on_load_failure_callback=pluginutils.log_fail_msg)
    return [[name, mgr[name].plugin] for name in mgr.names()]


_environment = None


def global_env():
    if _environment is None:
        initialise()
    return _environment


def initialise():
    global _environment
    if _environment is not None:
        return

    clients.initialise()

    global_env = environment.Environment({}, user_env=False)
    _load_global_environment(global_env)
    _environment = global_env
    global_env.registry.log_resource_info(show_all=True)


def _load_global_environment(env):
    _load_global_resources(env)
    environment.read_global_environment(env)


def _load_global_resources(env):
    _register_constraints(env, _get_mapping('heat.constraints'))
    _register_stack_lifecycle_plugins(
        env,
        _get_mapping('heat.stack_lifecycle_plugins'))
    _register_event_sinks(
        env,
        _get_mapping('heat.event_sinks'))

    manager = plugin_manager.PluginManager(__name__)
    # Sometimes resources should not be available for registration in Heat due
    # to unsatisfied dependencies. We look first for the function
    # 'available_resource_mapping', which should return the filtered resources.
    # If it is not found, we look for the legacy 'resource_mapping'.
    resource_mapping = plugin_manager.PluginMapping(['available_resource',
                                                     'resource'])
    constraint_mapping = plugin_manager.PluginMapping('constraint')

    _register_resources(env, resource_mapping.load_all(manager))

    _register_constraints(env, constraint_mapping.load_all(manager))


def list_opts():
    from heat.engine.resources.aws.lb import loadbalancer
    yield None, loadbalancer.loadbalancer_opts



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception as heat_exception
from heat.engine.clients import client_plugin
from heat.engine import constraints
from manilaclient import client as manila_client
from manilaclient import exceptions
from oslo_config import cfg

MANILACLIENT_VERSION = "2.13"
CLIENT_NAME = 'manila'


class ManilaClientPlugin(client_plugin.ClientPlugin):

    exceptions_module = exceptions

    service_types = [SHARE] = ['sharev2']

    def _create(self):
        endpoint_type = self._get_client_option(CLIENT_NAME, 'endpoint_type')
        args = {
            'endpoint_type': endpoint_type,
            'service_type': self.SHARE,
            'session': self.context.keystone_session,
            'connect_retries': cfg.CONF.client_retry_limit,
            'region_name': self._get_region_name()
        }
        client = manila_client.Client(MANILACLIENT_VERSION, **args)
        return client

    def is_not_found(self, ex):
        return isinstance(ex, exceptions.NotFound)

    def is_over_limit(self, ex):
        return isinstance(ex, exceptions.RequestEntityTooLarge)

    def is_conflict(self, ex):
        return isinstance(ex, exceptions.Conflict)

    @staticmethod
    def _find_resource_by_id_or_name(id_or_name, resource_list,
                                     resource_type_name):
        """The method is trying to find id or name in item_list

        The method searches item with id_or_name in list and returns it.
        If there is more than one value or no values then it raises an
        exception

        :param id_or_name: resource id or name
        :param resource_list: list of resources
        :param resource_type_name: name of resource type that will be used
                                   for exceptions
        :raises EntityNotFound: if cannot find resource by name
        :raises NoUniqueMatch: if find more than one resource by ambiguous name
        :return: resource or generate an exception otherwise
        """
        search_result_by_id = [res for res in resource_list
                               if res.id == id_or_name]
        if search_result_by_id:
            return search_result_by_id[0]
        else:
            # try to find resource by name
            search_result_by_name = [res for res in resource_list
                                     if res.name == id_or_name]
            match_count = len(search_result_by_name)
            if match_count > 1:
                message = ("Ambiguous {0} name '{1}'. Found more than one "
                           "{0} for this name in Manila."
                           ).format(resource_type_name, id_or_name)
                raise exceptions.NoUniqueMatch(message)
            elif match_count == 1:
                return search_result_by_name[0]
            else:
                raise heat_exception.EntityNotFound(entity=resource_type_name,
                                                    name=id_or_name)

    def get_share_type(self, share_type_identity):
        return self._find_resource_by_id_or_name(
            share_type_identity,
            self.client().share_types.list(),
            "share type"
        )

    def get_share_network(self, share_network_identity):
        return self._find_resource_by_id_or_name(
            share_network_identity,
            self.client().share_networks.list(),
            "share network"
        )

    def get_share_snapshot(self, snapshot_identity):
        return self._find_resource_by_id_or_name(
            snapshot_identity,
            self.client().share_snapshots.list(),
            "share snapshot"
        )

    def get_security_service(self, service_identity):
        return self._find_resource_by_id_or_name(
            service_identity,
            self.client().security_services.list(),
            'security service'
        )


class ManilaShareBaseConstraint(constraints.BaseCustomConstraint):

    # check that exceptions module has been loaded. Without this check
    # doc tests on gates will fail
    expected_exceptions = (heat_exception.EntityNotFound,
                           exceptions.NoUniqueMatch)
    resource_client_name = CLIENT_NAME


class ManilaShareNetworkConstraint(ManilaShareBaseConstraint):

    resource_getter_name = 'get_share_network'


class ManilaShareTypeConstraint(ManilaShareBaseConstraint):

    resource_getter_name = 'get_share_type'


class ManilaShareSnapshotConstraint(ManilaShareBaseConstraint):

    resource_getter_name = 'get_share_snapshot'



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from neutronclient.neutron import v2_0 as neutronV20

from heat.common import exception
from heat.common import template_format
from heat.engine import resource
from heat.engine.resources.openstack.magnum import cluster_template
from heat.engine import scheduler
from heat.engine import template
from heat.tests import common
from heat.tests import utils


RESOURCE_TYPE = 'OS::Magnum::ClusterTemplate'


class TestMagnumClusterTemplate(common.HeatTestCase):
    magnum_template = '''
    heat_template_version: ocata
    resources:
      test_cluster_template:
        type: OS::Magnum::ClusterTemplate
        properties:
          name: test_cluster_template
          image: fedora-21-atomic-2
          flavor: m1.small
          master_flavor: m1.medium
          keypair: heat_key
          external_network: 0244b54d-ae1f-44f0-a24a-442760f1d681
          fixed_network: 0f59a3dd-fac1-4d03-b41a-d4115fbffa89
          fixed_subnet: 27a8c89c-0d28-4946-8c78-82cfec1d670a
          dns_nameserver: 8.8.8.8
          docker_volume_size: 5
          docker_storage_driver: devicemapper
          coe: 'mesos'
          network_driver: 'flannel'
          http_proxy: 'http://proxy.com:123'
          https_proxy: 'https://proxy.com:123'
          no_proxy: '192.168.0.1'
          labels: {'flannel_cidr': ['10.101.0.0/16', '10.102.0.0/16']}
          tls_disabled: True
          public: True
          registry_enabled: True
          volume_driver: rexray
          server_type: vm
          master_lb_enabled: True
          floating_ip_enabled: True
    '''
    expected = {
        'name': 'test_cluster_template',
        'image_id': 'fedora-21-atomic-2',
        'flavor_id': 'm1.small',
        'master_flavor_id': 'm1.medium',
        'keypair_id': 'heat_key',
        'external_network_id': 'id_for_net_or_sub',
        'fixed_network': 'id_for_net_or_sub',
        'fixed_subnet': 'id_for_net_or_sub',
        'dns_nameserver': '8.8.8.8',
        'docker_volume_size': 5,
        'docker_storage_driver': 'devicemapper',
        'coe': 'mesos',
        'network_driver': 'flannel',
        'http_proxy': 'http://proxy.com:123',
        'https_proxy': 'https://proxy.com:123',
        'no_proxy': '192.168.0.1',
        'labels': {'flannel_cidr': ['10.101.0.0/16', '10.102.0.0/16']},
        'tls_disabled': True,
        'public': True,
        'registry_enabled': True,
        'volume_driver': 'rexray',
        'server_type': 'vm',
        'master_lb_enabled': True,
        'floating_ip_enabled': True
    }

    def setUp(self):
        super(TestMagnumClusterTemplate, self).setUp()
        resource._register_class(RESOURCE_TYPE,
                                 cluster_template.ClusterTemplate)
        self.t = template_format.parse(self.magnum_template)
        self.stack = utils.parse_stack(self.t)

        resource_defns = self.stack.t.resource_definitions(self.stack)
        self.rsrc_defn = resource_defns['test_cluster_template']
        self.client = mock.Mock()
        self.patchobject(cluster_template.ClusterTemplate, 'client',
                         return_value=self.client)
        self.find_mock = self.patchobject(neutronV20,
                                          'find_resourceid_by_name_or_id')
        self.find_mock.return_value = 'id_for_net_or_sub'
        self.stub_FlavorConstraint_validate()
        self.stub_KeypairConstraint_validate()
        self.stub_ImageConstraint_validate()
        self.stub_NetworkConstraint_validate()
        self.stub_SubnetConstraint_validate()

    def _create_resource(self, name, snippet, stack):
        self.resource_id = '12345'
        self.test_cluster_template = self.stack['test_cluster_template']
        value = mock.MagicMock(uuid=self.resource_id)
        self.client.cluster_templates.create.return_value = value
        bm = cluster_template.ClusterTemplate(name, snippet, stack)
        scheduler.TaskRunner(bm.create)()
        return bm

    def test_cluster_template_create(self):
        bm = self._create_resource('bm', self.rsrc_defn, self.stack)
        self.assertEqual(self.resource_id, bm.resource_id)
        self.assertEqual((bm.CREATE, bm.COMPLETE), bm.state)
        self.client.cluster_templates.create.assert_called_once_with(
            **self.expected)

    def test_validate_invalid_volume_driver(self):
        props = self.t['resources']['test_cluster_template']['properties']
        props['volume_driver'] = 'cinder'
        stack = utils.parse_stack(self.t)
        msg = ("Volume driver type cinder is not supported by COE:mesos, "
               "expecting a ['rexray'] volume driver.")
        ex = self.assertRaises(exception.StackValidationFailed,
                               stack['test_cluster_template'].validate)
        self.assertEqual(msg, str(ex))

    def _cluster_template_update(self, update_status='UPDATE_COMPLETE',
                                 exc_msg=None):
        ct = self._create_resource('ct', self.rsrc_defn, self.stack)
        status = mock.MagicMock(status=update_status)
        self.client.cluster_templates.get.return_value = status
        t = template_format.parse(self.magnum_template)
        new_t = copy.deepcopy(t)
        new_t['resources'][self.expected['name']]['properties'][
            cluster_template.ClusterTemplate.PUBLIC] = False
        rsrc_defns = template.Template(new_t).resource_definitions(self.stack)
        new_ct = rsrc_defns[self.expected['name']]
        if update_status == 'UPDATE_COMPLETE':
            scheduler.TaskRunner(ct.update, new_ct)()
            self.assertEqual((ct.UPDATE, ct.COMPLETE), ct.state)
        else:
            exc = self.assertRaises(
                exception.ResourceFailure,
                scheduler.TaskRunner(ct.update, new_ct))
            self.assertIn(exc_msg, str(exc))

    def test_cluster_update(self):
        self._cluster_template_update()

    def test_cluster_update_failed(self):
        self._cluster_template_update('UPDATE_FAILED',
                                      'Failed to update Cluster')

    def test_cluster_update_unknown_status(self):
        self._cluster_template_update('UPDATE_BAR',
                                      'Unknown status updating Cluster')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.engine.resources.openstack.neutron.taas import tap_flow
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

sample_template = {
    'heat_template_version': '2016-04-08',
    'resources': {
        'test_resource': {
            'type': 'OS::Neutron::TaaS::TapFlow',
            'properties': {
                'name': 'test_tap_flow',
                'description': 'desc',
                'port': '6af055d3-26f6-48dd-a597-7611d7e58d35',
                'tap_service': '6af055d3-26f6-48dd-a597-7611d7e58d35',
                'direction': 'BOTH',
                'vlan_filter': '1-5,9,18,27-30,99-108,4000-4095'
            }
        }
    }
}

RESOURCE_TYPE = 'OS::Neutron::TaaS::TapFlow'


class TapFlowTest(common.HeatTestCase):

    def setUp(self):
        super(TapFlowTest, self).setUp()

        self.ctx = utils.dummy_context()

        self.stack = stack.Stack(
            self.ctx, 'test_stack',
            template.Template(sample_template)
        )
        self.test_resource = self.stack['test_resource']

        self.test_client_plugin = mock.MagicMock()
        self.test_resource.client_plugin = mock.MagicMock(
            return_value=self.test_client_plugin)

        self.test_client = mock.MagicMock()
        self.test_resource.client = mock.MagicMock(
            return_value=self.test_client)

        self.test_client_plugin.get_notification = mock.MagicMock(
            return_value='sample_notification')

    def test_resource_mapping(self):
        mapping = tap_flow.resource_mapping()
        self.assertEqual(
            tap_flow.TapFlow,
            mapping['OS::Neutron::TaaS::TapFlow'])

    def _get_mock_resource(self):
        value = mock.MagicMock()
        value.id = '477e8273-60a7-4c41-b683-fdb0bc7cd152'
        return value

    def test_resource_handle_create(self):
        mock_tap_flow_create = self.test_client_plugin.create_ext_resource
        mock_resource = self._get_mock_resource()
        mock_tap_flow_create.return_value = mock_resource

    # validate the properties
        self.assertEqual(
            'test_tap_flow',
            self.test_resource.properties.get(
                tap_flow.TapFlow.NAME))
        self.assertEqual(
            'desc',
            self.test_resource.properties.get(
                tap_flow.TapFlow.DESCRIPTION))
        self.assertEqual(
            '6af055d3-26f6-48dd-a597-7611d7e58d35',
            self.test_resource.properties.get(
                tap_flow.TapFlow.PORT))
        self.assertEqual(
            '6af055d3-26f6-48dd-a597-7611d7e58d35',
            self.test_resource.properties.get(
                tap_flow.TapFlow.TAP_SERVICE))
        self.assertEqual(
            'BOTH',
            self.test_resource.properties.get(
                tap_flow.TapFlow.DIRECTION))
        self.assertEqual(
            '1-5,9,18,27-30,99-108,4000-4095',
            self.test_resource.properties.get(
                tap_flow.TapFlow.VLAN_FILTER))

        self.test_resource.data_set = mock.Mock()
        self.test_resource.handle_create()
        mock_tap_flow_create.assert_called_once_with(
            'tap_flow',
            {
                'name': 'test_tap_flow',
                'description': 'desc',
                'port': '6af055d3-26f6-48dd-a597-7611d7e58d35',
                'tap_service': '6af055d3-26f6-48dd-a597-7611d7e58d35',
                'direction': 'BOTH',
                'vlan_filter': '1-5,9,18,27-30,99-108,4000-4095',
            }
        )

    def test_resource_handle_delete(self):
        mock_tap_flow_delete = self.test_client_plugin.delete_ext_resource
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        mock_tap_flow_delete.return_value = None
        self.assertIsNone(self.test_resource.handle_delete())
        mock_tap_flow_delete.assert_called_once_with(
            'tap_flow', self.test_resource.resource_id)

    def test_resource_handle_delete_resource_id_is_none(self):
        self.test_resource.resource_id = None
        self.assertIsNone(self.test_resource.handle_delete())
        self.assertEqual(0, self.test_client_plugin.
                         delete_ext_resource.call_count)

    def test_resource_handle_delete_not_found(self):
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        mock_tap_flow_delete = self.test_client_plugin.delete_ext_resource
        mock_tap_flow_delete.side_effect = self.test_client_plugin.NotFound
        self.assertIsNone(self.test_resource.handle_delete())

    def test_resource_show_resource(self):
        mock_tap_flow_get = self.test_client_plugin.show_ext_resource
        mock_tap_flow_get.return_value = {}
        self.assertEqual({},
                         self.test_resource._show_resource(),
                         'Failed to show resource')

    def test_resource_handle_update(self):
        mock_tap_flow_patch = self.test_client_plugin.update_ext_resource
        self.test_resource.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        prop_diff = {
            tap_flow.TapFlow.NAME:
                'name-updated',
            tap_flow.TapFlow.DESCRIPTION:
                'description-updated',
            tap_flow.TapFlow.PORT:
                '6af055d3-26f6-48dd-a597-7611d7e58d35',
            tap_flow.TapFlow.TAP_SERVICE:
                '6af055d3-26f6-48dd-a597-7611d7e58d35',
            tap_flow.TapFlow.DIRECTION:
                'BOTH',
            tap_flow.TapFlow.VLAN_FILTER:
                '1-5,9,18,27-30,99-108,4000-4095'
            }
        self.test_resource.handle_update(json_snippet=None,
                                         tmpl_diff=None,
                                         prop_diff=prop_diff)

        mock_tap_flow_patch.assert_called_once_with(
            'tap_flow',
            {
                'name': 'name-updated',
                'description': 'description-updated',
                'port': '6af055d3-26f6-48dd-a597-7611d7e58d35',
                'tap_service': '6af055d3-26f6-48dd-a597-7611d7e58d35',
                'direction': 'BOTH',
                'vlan_filter': '1-5,9,18,27-30,99-108,4000-4095',
            }, self.test_resource.resource_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import mistral as mistral_client_plugin
from heat.engine import resource
from heat.engine import scheduler
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

from oslo_serialization import jsonutils

try:
    from zaqarclient.transport.errors import ResourceNotFound  # noqa
except ImportError:
    class ResourceNotFound(Exception):
        pass

subscr_template = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Resources" : {
    "MyQueue2" : {
      "Type" : "OS::Zaqar::Queue",
      "Properties" : {
        "name" : "myqueue",
        "metadata" : { "key1" : { "key2" : "value", "key3" : [1, 2] } }
      }
    },
    "MySubscription" : {
      "Type" : "OS::Zaqar::Subscription",
      "Properties" : {
        "queue_name" : "myqueue",
        "subscriber" : "mailto:name@domain.com",
        "ttl" : "3600",
        "options" : { "key1" : "value1" }
      }
    }
  }
}
'''

mistral_template = '''
{
  "heat_template_version" : "2015-10-15",
  "resources" : {
    "subscription" : {
      "type" : "OS::Zaqar::MistralTrigger",
      "properties" : {
        "queue_name" : "myqueue",
        "workflow_id": "abcd",
        "input" : { "key1" : "value1" }
      }
    }
  }
}
'''


class FakeSubscription(object):
    def __init__(self, queue_name, id=None, ttl=None, subscriber=None,
                 options=None, auto_create=True):
        self.id = id
        self.queue_name = queue_name
        self.ttl = ttl
        self.subscriber = subscriber
        self.options = options

    def update(self, prop_diff):
        allowed_keys = {'subscriber', 'ttl', 'options'}
        for key in prop_diff.keys():
            if key not in allowed_keys:
                raise KeyError(key)

    def delete(self):
        self._deleted = True


class ZaqarSubscriptionTest(common.HeatTestCase):
    def setUp(self):
        super(ZaqarSubscriptionTest, self).setUp()
        self.fc = mock.Mock()
        self.patchobject(resource.Resource, 'client', return_value=self.fc)
        self.ctx = utils.dummy_context()
        self.subscr_id = "58138648c1e2eb7355d62137"

    def parse_stack(self, t):
        stack_name = 'test_stack'
        tmpl = template.Template(t)
        self.stack = stack.Stack(self.ctx, stack_name, tmpl)
        self.stack.validate()
        self.stack.store()

    def test_validate_subscriber_type(self):
        t = template_format.parse(subscr_template)
        t['Resources']['MySubscription']['Properties']['subscriber'] = "foo:ba"
        stack_name = 'test_stack'
        tmpl = template.Template(t)
        self.stack = stack.Stack(self.ctx, stack_name, tmpl)

        exc = self.assertRaises(exception.StackValidationFailed,
                                self.stack.validate)
        self.assertEqual('The subscriber type of must be one of: http, https, '
                         'mailto, trust+http, trust+https.',
                         str(exc))

    def test_create(self):
        t = template_format.parse(subscr_template)
        self.parse_stack(t)

        subscr = self.stack['MySubscription']

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        self.fc.subscription.return_value = fake_subscr

        scheduler.TaskRunner(subscr.create)()
        self.assertEqual(self.subscr_id, subscr.FnGetRefId())
        self.fc.subscription.assert_called_once_with(
            subscr.properties['queue_name'],
            options={'key1': 'value1'},
            subscriber='mailto:name@domain.com',
            ttl=3600)

    def test_delete(self):
        t = template_format.parse(subscr_template)
        self.parse_stack(t)

        subscr = self.stack['MySubscription']

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        self.fc.subscription.return_value = fake_subscr
        scheduler.TaskRunner(subscr.create)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options={'key1': 'value1'},
            subscriber='mailto:name@domain.com',
            ttl=3600)
        scheduler.TaskRunner(subscr.delete)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            id=self.subscr_id,
            auto_create=False)
        self.assertTrue(getattr(fake_subscr, '_deleted', False))

    def test_delete_not_found(self):
        t = template_format.parse(subscr_template)
        self.parse_stack(t)

        subscr = self.stack['MySubscription']

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        self.fc.subscription.side_effect = [fake_subscr, ResourceNotFound]

        scheduler.TaskRunner(subscr.create)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options={'key1': 'value1'},
            subscriber='mailto:name@domain.com',
            ttl=3600)
        scheduler.TaskRunner(subscr.delete)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            id=self.subscr_id,
            auto_create=False)

    def test_update_in_place(self):
        t = template_format.parse(subscr_template)
        self.parse_stack(t)

        subscr = self.stack['MySubscription']

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        self.fc.subscription.return_value = fake_subscr
        t = template_format.parse(subscr_template)
        new_subscr = t['Resources']['MySubscription']
        new_subscr['Properties']['ttl'] = "3601"
        resource_defns = template.Template(t).resource_definitions(self.stack)

        scheduler.TaskRunner(subscr.create)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options={'key1': 'value1'},
            subscriber='mailto:name@domain.com',
            ttl=3600)
        scheduler.TaskRunner(subscr.update, resource_defns['MySubscription'])()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            id=self.subscr_id,
            auto_create=False)

    def test_update_replace(self):
        t = template_format.parse(subscr_template)
        self.parse_stack(t)

        subscr = self.stack['MySubscription']

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        self.fc.subscription.return_value = fake_subscr

        t = template_format.parse(subscr_template)
        t['Resources']['MySubscription']['Properties']['queue_name'] = 'foo'
        resource_defns = template.Template(t).resource_definitions(self.stack)
        new_subscr = resource_defns['MySubscription']

        scheduler.TaskRunner(subscr.create)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options={'key1': 'value1'},
            subscriber='mailto:name@domain.com',
            ttl=3600)
        err = self.assertRaises(resource.UpdateReplace,
                                scheduler.TaskRunner(subscr.update,
                                                     new_subscr))
        msg = 'The Resource MySubscription requires replacement.'
        self.assertEqual(msg, str(err))

    def test_show_resource(self):
        t = template_format.parse(subscr_template)
        self.parse_stack(t)

        subscr = self.stack['MySubscription']

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        props = t['Resources']['MySubscription']['Properties']
        fake_subscr.ttl = props['ttl']
        fake_subscr.subscriber = props['subscriber']
        fake_subscr.options = props['options']

        self.fc.subscription.return_value = fake_subscr

        rsrc_data = props.copy()
        rsrc_data['id'] = self.subscr_id
        scheduler.TaskRunner(subscr.create)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options={'key1': 'value1'},
            subscriber='mailto:name@domain.com',
            ttl=3600)
        self.assertEqual(rsrc_data, subscr._show_resource())
        self.assertEqual(
            {'queue_name': props['queue_name'],
             'subscriber': props['subscriber'],
             'ttl': props['ttl'],
             'options': props['options']},
            subscr.parse_live_resource_data(subscr.properties,
                                            subscr._show_resource()))

        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            id=self.subscr_id,
            auto_create=False)


class JsonString(object):
    def __init__(self, data):
        self._data = data

    def __eq__(self, other):
        return self._data == jsonutils.loads(other)

    def __str__(self):
        return jsonutils.dumps(self._data)

    def __repr__(self):
        return str(self)


class ZaqarMistralTriggerTest(common.HeatTestCase):
    def setUp(self):
        super(ZaqarMistralTriggerTest, self).setUp()
        self.fc = mock.Mock()
        self.patchobject(resource.Resource, 'client', return_value=self.fc)
        self.ctx = utils.dummy_context()
        self.patchobject(mistral_client_plugin.WorkflowConstraint,
                         'validate', return_value=True)
        self.subscr_id = "58138648c1e2eb7355d62137"

        stack_name = 'test_stack'
        t = template_format.parse(mistral_template)
        tmpl = template.Template(t)
        self.stack = stack.Stack(self.ctx, stack_name, tmpl)
        self.stack.validate()
        self.stack.store()

        def client(name='zaqar'):
            if name == 'mistral':
                client = mock.Mock()
                http_client = mock.Mock()
                client.executions = mock.Mock(spec=['http_client'])
                client.executions.http_client = http_client
                http_client.base_url = 'http://mistral.example.net:8989'
                return client
            elif name == 'zaqar':
                return self.fc

        self.subscr = self.stack['subscription']
        self.subscr.client = mock.Mock(side_effect=client)

        self.subscriber = 'trust+http://mistral.example.net:8989/executions'
        self.options = {
            'post_data': JsonString({
                'workflow_id': 'abcd',
                'input': {"key1": "value1"},
                'params': {'env': {'notification': '$zaqar_message$'}},
            })
        }

    def test_create(self):
        subscr = self.subscr

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        self.fc.subscription.return_value = fake_subscr

        scheduler.TaskRunner(subscr.create)()
        self.assertEqual(self.subscr_id, subscr.FnGetRefId())
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options=self.options,
            subscriber=self.subscriber,
            ttl=220367260800)

    def test_delete(self):
        subscr = self.subscr

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        self.fc.subscription.return_value = fake_subscr
        scheduler.TaskRunner(subscr.create)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options=self.options,
            subscriber=self.subscriber,
            ttl=220367260800)
        scheduler.TaskRunner(subscr.delete)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            id=self.subscr_id,
            auto_create=False)
        self.assertTrue(getattr(fake_subscr, '_deleted', False))

    def test_delete_not_found(self):
        subscr = self.subscr

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        self.fc.subscription.side_effect = [fake_subscr, ResourceNotFound]
        scheduler.TaskRunner(subscr.create)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options=self.options,
            subscriber=self.subscriber,
            ttl=220367260800)
        scheduler.TaskRunner(subscr.delete)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            id=self.subscr_id,
            auto_create=False)

    def test_update_in_place(self):
        subscr = self.subscr

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        self.fc.subscription.return_value = fake_subscr

        t = template_format.parse(mistral_template)
        new_subscr = t['resources']['subscription']
        new_subscr['properties']['ttl'] = "3601"
        resource_defns = template.Template(t).resource_definitions(self.stack)

        scheduler.TaskRunner(subscr.create)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options=self.options,
            subscriber=self.subscriber,
            ttl=220367260800)
        scheduler.TaskRunner(subscr.update, resource_defns['subscription'])()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            id=self.subscr_id,
            auto_create=False)

    def test_update_replace(self):
        subscr = self.subscr

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        self.fc.subscription.return_value = fake_subscr

        t = template_format.parse(mistral_template)
        t['resources']['subscription']['properties']['queue_name'] = 'foo'
        resource_defns = template.Template(t).resource_definitions(self.stack)
        new_subscr = resource_defns['subscription']

        scheduler.TaskRunner(subscr.create)()
        err = self.assertRaises(resource.UpdateReplace,
                                scheduler.TaskRunner(subscr.update,
                                                     new_subscr))
        msg = 'The Resource subscription requires replacement.'
        self.assertEqual(msg, str(err))
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options=self.options,
            subscriber=self.subscriber,
            ttl=220367260800)

    def test_show_resource(self):
        subscr = self.subscr

        fake_subscr = FakeSubscription(subscr.properties['queue_name'],
                                       self.subscr_id)
        fake_subscr.ttl = 220367260800
        fake_subscr.subscriber = self.subscriber
        fake_subscr.options = {'post_data': str(self.options['post_data'])}

        self.fc.subscription.return_value = fake_subscr

        props = self.stack.t.t['resources']['subscription']['properties']
        scheduler.TaskRunner(subscr.create)()
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            options=self.options,
            subscriber=self.subscriber,
            ttl=220367260800)
        self.assertEqual(
            {'queue_name': props['queue_name'],
             'id': self.subscr_id,
             'subscriber': self.subscriber,
             'options': self.options,
             'ttl': 220367260800},
            subscr._show_resource())
        self.assertEqual(
            {'queue_name': props['queue_name'],
             'workflow_id': props['workflow_id'],
             'input': props['input'],
             'params': {},
             'ttl': 220367260800},
            subscr.parse_live_resource_data(subscr.properties,
                                            subscr._show_resource()))
        self.fc.subscription.assert_called_with(
            subscr.properties['queue_name'],
            id=self.subscr_id,
            auto_create=False)



#
# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from webob import exc

from heat.common import endpoint_utils
from heat.common.i18n import _
from heat.common import wsgi


class AuthUrlFilter(wsgi.Middleware):

    def __init__(self, app, conf):
        super(AuthUrlFilter, self).__init__(app)
        self.conf = conf
        self._auth_url = None

    @property
    def auth_url(self):
        if not self._auth_url:
            self._auth_url = self._get_auth_url()
        return self._auth_url

    def _get_auth_url(self):
        if 'auth_uri' in self.conf:
            return self.conf['auth_uri']
        else:
            return endpoint_utils.get_auth_uri(v3=False)

    def _validate_auth_url(self, auth_url):
        """Validate auth_url to ensure it can be used."""
        if not auth_url:
            raise exc.HTTPBadRequest(_('Request missing required header '
                                       'X-Auth-Url'))
        allowed = cfg.CONF.auth_password.allowed_auth_uris
        if auth_url not in allowed:
            raise exc.HTTPUnauthorized(_('Header X-Auth-Url "%s" not '
                                         'an allowed endpoint') % auth_url)
        return True

    def process_request(self, req):
        auth_url = self.auth_url
        if cfg.CONF.auth_password.multi_cloud:
            auth_url = req.headers.get('X-Auth-Url')
            self._validate_auth_url(auth_url)

        req.headers['X-Auth-Url'] = auth_url
        return None


def filter_factory(global_conf, **local_conf):
    conf = global_conf.copy()
    conf.update(local_conf)

    def auth_url_filter(app):
        return AuthUrlFilter(app, conf)
    return auth_url_filter



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from heat.common import exception
from heat.common.i18n import _

LOG = logging.getLogger(__name__)

PARAMETER_GROUPS = 'parameter_groups'
PARAMETERS = 'parameters'


class ParameterGroups(object):
    """The ParameterGroups specified by the stack's template."""
    def __init__(self, tmpl):
        self.tmpl = tmpl
        self.parameters = tmpl.parameters(None, {}, param_defaults={})
        self.parameter_names = []
        if self.parameters:
            self.parameter_names = [param for param in self.parameters]
        self.parameter_groups = tmpl.get(PARAMETER_GROUPS)

    def validate(self):
        """Validate the parameter group.

        Validate that each parameter belongs to only one Parameter Group and
        that each parameter name in the group references a valid parameter.
        """
        LOG.debug('Validating Parameter Groups: %s',
                  ', '.join(self.parameter_names))
        if self.parameter_groups:
            if not isinstance(self.parameter_groups, list):
                raise exception.StackValidationFailed(
                    error=_('Parameter Groups error'),
                    path=[PARAMETER_GROUPS],
                    message=_('The %s should be a list.') % PARAMETER_GROUPS)

            # Loop through groups and validate parameters
            grouped_parameters = []
            for group in self.parameter_groups:
                parameters = group.get(PARAMETERS)
                if parameters is None:
                    raise exception.StackValidationFailed(
                        error=_('Parameter Groups error'),
                        path=[PARAMETER_GROUPS, group.get('label', '')],
                        message=_('The %s must be provided for '
                                  'each parameter group.') % PARAMETERS)

                if not isinstance(parameters, list):
                    raise exception.StackValidationFailed(
                        error=_('Parameter Groups error'),
                        path=[PARAMETER_GROUPS, group.get('label', '')],
                        message=_('The %s of parameter group '
                                  'should be a list.') % PARAMETERS)

                for param in parameters:
                    # Check if param has been added to a previous group
                    if param in grouped_parameters:
                        raise exception.StackValidationFailed(
                            error=_('Parameter Groups error'),
                            path=[PARAMETER_GROUPS, group.get('label', '')],
                            message=_(
                                'The %s parameter must be assigned to one '
                                'parameter group only.') % param)
                    else:
                        grouped_parameters.append(param)

                    # Check that grouped parameter references a valid Parameter
                    if param not in self.parameter_names:
                        raise exception.StackValidationFailed(
                            error=_('Parameter Groups error'),
                            path=[PARAMETER_GROUPS, group.get('label', '')],
                            message=_(
                                'The grouped parameter %s does not reference '
                                'a valid parameter.') % param)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.engine.resources.openstack.senlin import res_base


class Node(res_base.BaseSenlinResource):
    """A resource that creates a Senlin Node.

    Node is an object that belongs to at most one Cluster, it can be created
    based on a profile.
    """
    pass


def resource_mapping():
    return {
        'OS::Senlin::Node': Node
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from heat_integrationtests.functional import functional_base


class DeleteInProgressTest(functional_base.FunctionalTestsBase):

    root_template = '''
heat_template_version: 2013-05-23
resources:
    rg:
        type: OS::Heat::ResourceGroup
        properties:
            count: 125
            resource_def:
                type: empty.yaml
'''

    empty_template = '''
heat_template_version: 2013-05-23
resources:
'''

    def test_delete_nested_stacks_create_in_progress(self):
        files = {'empty.yaml': self.empty_template}
        identifier = self.stack_create(template=self.root_template,
                                       files=files,
                                       expected_status='CREATE_IN_PROGRESS')
        time.sleep(20)
        self._stack_delete(identifier)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support


class CinderEncryptedVolumeType(resource.Resource):
    """A resource for encrypting a cinder volume type.

    A Volume Encryption Type is a collection of settings used to conduct
    encryption for a specific volume type.

    Note that default cinder security policy usage of this resource
    is limited to being used by administrators only.
    """

    support_status = support.SupportStatus(version='5.0.0')

    default_client_name = 'cinder'

    entity = 'volume_encryption_types'

    required_service_extension = 'encryption'

    PROPERTIES = (
        PROVIDER, CONTROL_LOCATION, CIPHER, KEY_SIZE, VOLUME_TYPE
    ) = (
        'provider', 'control_location', 'cipher', 'key_size', 'volume_type'
    )

    properties_schema = {
        PROVIDER: properties.Schema(
            properties.Schema.STRING,
            _('The class that provides encryption support. '
              'For example, nova.volume.encryptors.luks.LuksEncryptor.'),
            required=True,
            update_allowed=True
        ),
        CONTROL_LOCATION: properties.Schema(
            properties.Schema.STRING,
            _('Notional service where encryption is performed '
              'For example, front-end. For Nova.'),
            constraints=[
                constraints.AllowedValues(['front-end', 'back-end'])
            ],
            default='front-end',
            update_allowed=True
        ),
        CIPHER: properties.Schema(
            properties.Schema.STRING,
            _('The encryption algorithm or mode. '
              'For example, aes-xts-plain64.'),
            constraints=[
                constraints.AllowedValues(
                    ['aes-xts-plain64', 'aes-cbc-essiv']
                )
            ],
            default=None,
            update_allowed=True
        ),
        KEY_SIZE: properties.Schema(
            properties.Schema.INTEGER,
            _('Size of encryption key, in bits. '
              'For example, 128 or 256.'),
            default=None,
            update_allowed=True
        ),
        VOLUME_TYPE: properties.Schema(
            properties.Schema.STRING,
            _('Name or id of volume type (OS::Cinder::VolumeType).'),
            required=True,
            constraints=[constraints.CustomConstraint('cinder.vtype')]
        ),
    }

    def _get_vol_type_id(self, volume_type):
        id = self.client_plugin().get_volume_type(volume_type)
        return id

    def handle_create(self):
        body = {
            'provider': self.properties[self.PROVIDER],
            'cipher': self.properties[self.CIPHER],
            'key_size': self.properties[self.KEY_SIZE],
            'control_location': self.properties[self.CONTROL_LOCATION]
        }

        vol_type_id = self._get_vol_type_id(self.properties[self.VOLUME_TYPE])

        encrypted_vol_type = self.client().volume_encryption_types.create(
            volume_type=vol_type_id, specs=body
        )
        self.resource_id_set(encrypted_vol_type.volume_type_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.client().volume_encryption_types.update(
                volume_type=self.resource_id, specs=prop_diff
            )

    def get_live_resource_data(self):
        try:
            resource_data = self._show_resource()
            if not resource_data:
                # use attribute error, e.g. API call get raises AttributeError,
                # when evt is not exists or not ready (cinder bug 1562024).
                raise AttributeError()
        except Exception as ex:
            if (self.client_plugin().is_not_found(ex) or
                    isinstance(ex, AttributeError)):
                raise exception.EntityNotFound(entity='Resource',
                                               name=self.name)
            raise
        return resource_data

    def parse_live_resource_data(self, resource_properties, resource_data):
        resource_reality = {}

        for key in set(self.PROPERTIES) - {self.VOLUME_TYPE}:
            resource_reality.update({key: resource_data.get(key)})

        return resource_reality


def resource_mapping():
    return {
        'OS::Cinder::EncryptedVolumeType': CinderEncryptedVolumeType
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import webob.exc

import heat.api.middleware.fault as fault
import heat.api.openstack.v1.resources as resources
from heat.common import exception as heat_exc
from heat.common import identifier
from heat.common import policy
from heat.rpc import api as rpc_api
from heat.rpc import client as rpc_client
from heat.tests.api.openstack_v1 import tools
from heat.tests import common


@mock.patch.object(policy.Enforcer, 'enforce')
class ResourceControllerTest(tools.ControllerTest, common.HeatTestCase):
    """Tests the API class ResourceController.

    Tests the API class which acts as the WSGI controller,
    the endpoint processing API requests after they are routed
    """

    def setUp(self):
        super(ResourceControllerTest, self).setUp()
        # Create WSGI controller instance

        class DummyConfig(object):
            bind_port = 8004

        cfgopts = DummyConfig()
        self.controller = resources.ResourceController(options=cfgopts)

    def test_index(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(stack_identity._tenant_path() + '/resources')

        engine_resp = [
            {
                'resource_identity': dict(res_identity),
                'stack_name': stack_identity.stack_name,
                'resource_name': res_name,
                'resource_status_reason': None,
                'updated_time': '2012-07-23T13:06:00Z',
                'stack_identity': stack_identity,
                'resource_action': 'CREATE',
                'resource_status': 'COMPLETE',
                'physical_resource_id':
                'a3455d8c-9f88-404d-a85b-5315293e67de',
                'resource_type': 'AWS::EC2::Instance',
            }
        ]
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=engine_resp)

        result = self.controller.index(req, tenant_id=self.tenant,
                                       stack_name=stack_identity.stack_name,
                                       stack_id=stack_identity.stack_id)

        expected = {
            'resources': [{'links': [{'href': self._url(res_identity),
                                      'rel': 'self'},
                                     {'href': self._url(stack_identity),
                                      'rel': 'stack'}],
                           'resource_name': res_name,
                           'logical_resource_id': res_name,
                           'resource_status_reason': None,
                           'updated_time': '2012-07-23T13:06:00Z',
                           'resource_status': 'CREATE_COMPLETE',
                           'physical_resource_id':
                           'a3455d8c-9f88-404d-a85b-5315293e67de',
                           'resource_type': 'AWS::EC2::Instance'}]}
        self.assertEqual(expected, result)

        mock_call.assert_called_once_with(
            req.context,
            ('list_stack_resources', {'stack_identity': stack_identity,
                                      'nested_depth': 0,
                                      'with_detail': False,
                                      'filters': {}
                                      }),
            version='1.25'
        )

    def test_index_nonexist(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', True)
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'rubbish', '1')

        req = self._get(stack_identity._tenant_path() + '/resources')

        error = heat_exc.EntityNotFound(entity='Stack', name='a')
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     side_effect=tools.to_remote_error(error))

        resp = tools.request_with_middleware(
            fault.FaultWrapper,
            self.controller.index,
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id)

        self.assertEqual(404, resp.json['code'])
        self.assertEqual('EntityNotFound', resp.json['error']['type'])

        mock_call.assert_called_once_with(
            req.context,
            ('list_stack_resources', {'stack_identity': stack_identity,
                                      'nested_depth': 0,
                                      'with_detail': False,
                                      'filters': {}}),
            version='1.25'
        )

    def test_index_invalid_filters(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', True)
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'rubbish', '1')

        req = self._get(stack_identity._tenant_path() + '/resources',
                        {'invalid_key': 'junk'})

        mock_call = self.patchobject(rpc_client.EngineClient, 'call')
        ex = self.assertRaises(webob.exc.HTTPBadRequest,
                               self.controller.index, req,
                               tenant_id=self.tenant,
                               stack_name=stack_identity.stack_name,
                               stack_id=stack_identity.stack_id)

        self.assertIn("Invalid filter parameters %s" %
                      [str('invalid_key')],
                      str(ex))
        self.assertFalse(mock_call.called)

    def test_index_nested_depth(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', True)
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'rubbish', '1')

        req = self._get(stack_identity._tenant_path() + '/resources',
                        {'nested_depth': '99'})

        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=[])

        result = self.controller.index(req, tenant_id=self.tenant,
                                       stack_name=stack_identity.stack_name,
                                       stack_id=stack_identity.stack_id)

        self.assertEqual([], result['resources'])
        mock_call.assert_called_once_with(
            req.context,
            ('list_stack_resources', {'stack_identity': stack_identity,
                                      'nested_depth': 99,
                                      'with_detail': False,
                                      'filters': {}}),
            version='1.25'
        )

    def test_index_nested_depth_not_int(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', True)
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'rubbish', '1')

        req = self._get(stack_identity._tenant_path() + '/resources',
                        {'nested_depth': 'non-int'})

        mock_call = self.patchobject(rpc_client.EngineClient, 'call')
        ex = self.assertRaises(webob.exc.HTTPBadRequest,
                               self.controller.index, req,
                               tenant_id=self.tenant,
                               stack_name=stack_identity.stack_name,
                               stack_id=stack_identity.stack_id)

        self.assertEqual("Only integer is acceptable by 'nested_depth'.",
                         str(ex))
        self.assertFalse(mock_call.called)

    def test_index_denied_policy(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', False)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')
        identifier.ResourceIdentifier(resource_name=res_name,
                                      **stack_identity)

        req = self._get(stack_identity._tenant_path() + '/resources')

        resp = tools.request_with_middleware(
            fault.FaultWrapper,
            self.controller.index,
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id)

        self.assertEqual(403, resp.status_int)
        self.assertIn('403 Forbidden', str(resp))

    def test_index_detail(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(stack_identity._tenant_path() + '/resources',
                        {'with_detail': 'true'})

        resp_parameters = {
            "OS::project_id": "3ab5b02fa01f4f95afa1e254afc4a435",
            "network": "cf05086d-07c7-4ed6-95e5-e4af724677e6",
            "OS::stack_name": "s1", "admin_pass": "******",
            "key_name": "kk", "image": "fa5d387e-541f-4dfb-ae8a-83a614683f84",
            "db_port": "50000",
            "OS::stack_id": "723d7cee-46b3-4433-9c21-f3378eb0bfc4",
            "flavor": "1"
        },

        engine_resp = [
            {
                'resource_identity': dict(res_identity),
                'stack_name': stack_identity.stack_name,
                'resource_name': res_name,
                'resource_status_reason': None,
                'updated_time': '2012-07-23T13:06:00Z',
                'stack_identity': stack_identity,
                'resource_action': 'CREATE',
                'resource_status': 'COMPLETE',
                'physical_resource_id':
                'a3455d8c-9f88-404d-a85b-5315293e67de',
                'resource_type': 'AWS::EC2::Instance',
                'parameters': resp_parameters,
                'description': 'Hello description',
                'stack_user_project_id': '6f38bcfebbc4400b82d50c1a2ea3057d',
            }
        ]
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=engine_resp)

        result = self.controller.index(req, tenant_id=self.tenant,
                                       stack_name=stack_identity.stack_name,
                                       stack_id=stack_identity.stack_id)

        expected = {
            'resources': [{'links': [{'href': self._url(res_identity),
                                      'rel': 'self'},
                                     {'href': self._url(stack_identity),
                                      'rel': 'stack'}],
                           'resource_name': res_name,
                           'logical_resource_id': res_name,
                           'resource_status_reason': None,
                           'updated_time': '2012-07-23T13:06:00Z',
                           'resource_status': 'CREATE_COMPLETE',
                           'physical_resource_id':
                           'a3455d8c-9f88-404d-a85b-5315293e67de',
                           'resource_type': 'AWS::EC2::Instance',
                           'parameters': resp_parameters,
                           'description': 'Hello description',
                           'stack_user_project_id':
                           '6f38bcfebbc4400b82d50c1a2ea3057d'}]}
        self.assertEqual(expected, result)

        mock_call.assert_called_once_with(
            req.context,
            ('list_stack_resources', {'stack_identity': stack_identity,
                                      'nested_depth': 0,
                                      'with_detail': True,
                                      'filters': {}}),
            version='1.25'
        )

    def test_show(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'show', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '6')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(stack_identity._tenant_path())

        engine_resp = {
            'description': '',
            'resource_identity': dict(res_identity),
            'stack_name': stack_identity.stack_name,
            'resource_name': res_name,
            'resource_status_reason': None,
            'updated_time': '2012-07-23T13:06:00Z',
            'stack_identity': dict(stack_identity),
            'resource_action': 'CREATE',
            'resource_status': 'COMPLETE',
            'physical_resource_id':
            'a3455d8c-9f88-404d-a85b-5315293e67de',
            'resource_type': 'AWS::EC2::Instance',
            'attributes': {'foo': 'bar'},
            'metadata': {'ensureRunning': 'true'}
        }
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=engine_resp)

        result = self.controller.show(req, tenant_id=self.tenant,
                                      stack_name=stack_identity.stack_name,
                                      stack_id=stack_identity.stack_id,
                                      resource_name=res_name)

        expected = {
            'resource': {
                'links': [
                    {'href': self._url(res_identity), 'rel': 'self'},
                    {'href': self._url(stack_identity), 'rel': 'stack'},
                ],
                'description': '',
                'resource_name': res_name,
                'logical_resource_id': res_name,
                'resource_status_reason': None,
                'updated_time': '2012-07-23T13:06:00Z',
                'resource_status': 'CREATE_COMPLETE',
                'physical_resource_id': 'a3455d8c-9f88-404d-a85b-5315293e67de',
                'resource_type': 'AWS::EC2::Instance',
                'attributes': {'foo': 'bar'},
            }
        }
        self.assertEqual(expected, result)

        mock_call.assert_called_once_with(
            req.context,
            ('describe_stack_resource',
             {'stack_identity': stack_identity, 'resource_name': res_name,
              'with_attr': None}),
            version='1.2'
        )

    def test_show_with_nested_stack(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'show', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '6')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)
        nested_stack_identity = identifier.HeatIdentifier(self.tenant,
                                                          'nested', 'some_id')

        req = self._get(stack_identity._tenant_path())

        engine_resp = {
            'description': '',
            'resource_identity': dict(res_identity),
            'stack_name': stack_identity.stack_name,
            'resource_name': res_name,
            'resource_status_reason': None,
            'updated_time': '2012-07-23T13:06:00Z',
            'stack_identity': dict(stack_identity),
            'resource_action': 'CREATE',
            'resource_status': 'COMPLETE',
            'physical_resource_id': 'a3455d8c-9f88-404d-a85b-5315293e67de',
            'resource_type': 'AWS::EC2::Instance',
            'attributes': {'foo': 'bar'},
            'metadata': {'ensureRunning': 'true'},
            'nested_stack_id': dict(nested_stack_identity)
        }
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=engine_resp)

        result = self.controller.show(req, tenant_id=self.tenant,
                                      stack_name=stack_identity.stack_name,
                                      stack_id=stack_identity.stack_id,
                                      resource_name=res_name)

        expected = [{'href': self._url(res_identity), 'rel': 'self'},
                    {'href': self._url(stack_identity), 'rel': 'stack'},
                    {'href': self._url(nested_stack_identity), 'rel': 'nested'}
                    ]
        self.assertEqual(expected, result['resource']['links'])
        self.assertIsNone(result.get(rpc_api.RES_NESTED_STACK_ID))

        mock_call.assert_called_once_with(
            req.context,
            ('describe_stack_resource',
             {'stack_identity': stack_identity, 'resource_name': res_name,
              'with_attr': None}),
            version='1.2'
        )

    def test_show_nonexist(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'show', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'rubbish', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(res_identity._tenant_path())

        error = heat_exc.EntityNotFound(entity='Stack', name='a')
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     side_effect=tools.to_remote_error(error))

        resp = tools.request_with_middleware(
            fault.FaultWrapper,
            self.controller.show,
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id,
            resource_name=res_name)

        self.assertEqual(404, resp.json['code'])
        self.assertEqual('EntityNotFound', resp.json['error']['type'])

        mock_call.assert_called_once_with(
            req.context,
            ('describe_stack_resource',
             {'stack_identity': stack_identity, 'resource_name': res_name,
              'with_attr': None}),
            version='1.2'
        )

    def test_show_with_single_attribute(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'show', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant, 'foo', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)
        mock_describe = mock.Mock(return_value={'foo': 'bar'})
        self.controller.rpc_client.describe_stack_resource = mock_describe

        req = self._get(res_identity._tenant_path(), {'with_attr': 'baz'})
        resp = self.controller.show(req, tenant_id=self.tenant,
                                    stack_name=stack_identity.stack_name,
                                    stack_id=stack_identity.stack_id,
                                    resource_name=res_name)

        self.assertEqual({'resource': {'foo': 'bar'}}, resp)
        args, kwargs = mock_describe.call_args
        self.assertIn('baz', kwargs['with_attr'])

    def test_show_with_multiple_attributes(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'show', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant, 'foo', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)
        mock_describe = mock.Mock(return_value={'foo': 'bar'})
        self.controller.rpc_client.describe_stack_resource = mock_describe

        req = self._get(res_identity._tenant_path())
        req.environ['QUERY_STRING'] = 'with_attr=a1&with_attr=a2&with_attr=a3'
        resp = self.controller.show(req, tenant_id=self.tenant,
                                    stack_name=stack_identity.stack_name,
                                    stack_id=stack_identity.stack_id,
                                    resource_name=res_name)

        self.assertEqual({'resource': {'foo': 'bar'}}, resp)
        args, kwargs = mock_describe.call_args
        self.assertIn('a1', kwargs['with_attr'])
        self.assertIn('a2', kwargs['with_attr'])
        self.assertIn('a3', kwargs['with_attr'])

    def test_show_nonexist_resource(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'show', True)
        res_name = 'Wibble'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(res_identity._tenant_path())

        error = heat_exc.ResourceNotFound(stack_name='a', resource_name='b')
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     side_effect=tools.to_remote_error(error))

        resp = tools.request_with_middleware(
            fault.FaultWrapper,
            self.controller.show,
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id,
            resource_name=res_name)

        self.assertEqual(404, resp.json['code'])
        self.assertEqual('ResourceNotFound', resp.json['error']['type'])

        mock_call.assert_called_once_with(
            req.context,
            ('describe_stack_resource',
             {'stack_identity': stack_identity, 'resource_name': res_name,
              'with_attr': None}),
            version='1.2'
        )

    def test_show_uncreated_resource(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'show', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(res_identity._tenant_path())

        error = heat_exc.ResourceNotAvailable(resource_name='')
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     side_effect=tools.to_remote_error(error))

        resp = tools.request_with_middleware(
            fault.FaultWrapper,
            self.controller.show,
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id,
            resource_name=res_name)

        self.assertEqual(404, resp.json['code'])
        self.assertEqual('ResourceNotAvailable', resp.json['error']['type'])

        mock_call.assert_called_once_with(
            req.context,
            ('describe_stack_resource',
             {'stack_identity': stack_identity, 'resource_name': res_name,
              'with_attr': None}),
            version='1.2'
        )

    def test_show_err_denied_policy(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'show', False)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(res_identity._tenant_path())

        resp = tools.request_with_middleware(
            fault.FaultWrapper,
            self.controller.show,
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id,
            resource_name=res_name)

        self.assertEqual(403, resp.status_int)
        self.assertIn('403 Forbidden', str(resp))

    def test_metadata_show(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'metadata', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '6')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(stack_identity._tenant_path())

        engine_resp = {
            'description': '',
            'resource_identity': dict(res_identity),
            'stack_name': stack_identity.stack_name,
            'resource_name': res_name,
            'resource_status_reason': None,
            'updated_time': '2012-07-23T13:06:00Z',
            'stack_identity': dict(stack_identity),
            'resource_action': 'CREATE',
            'resource_status': 'COMPLETE',
            'physical_resource_id': 'a3455d8c-9f88-404d-a85b-5315293e67de',
            'resource_type': 'AWS::EC2::Instance',
            'metadata': {'ensureRunning': 'true'}
        }
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=engine_resp)

        result = self.controller.metadata(req, tenant_id=self.tenant,
                                          stack_name=stack_identity.stack_name,
                                          stack_id=stack_identity.stack_id,
                                          resource_name=res_name)

        expected = {'metadata': {'ensureRunning': 'true'}}
        self.assertEqual(expected, result)

        mock_call.assert_called_once_with(
            req.context,
            ('describe_stack_resource',
             {'stack_identity': stack_identity, 'resource_name': res_name,
              'with_attr': False}),
            version='1.2'
        )

    def test_metadata_show_nonexist(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'metadata', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'rubbish', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(res_identity._tenant_path() + '/metadata')

        error = heat_exc.EntityNotFound(entity='Stack', name='a')
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     side_effect=tools.to_remote_error(error))

        resp = tools.request_with_middleware(
            fault.FaultWrapper,
            self.controller.metadata,
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id,
            resource_name=res_name)

        self.assertEqual(404, resp.json['code'])
        self.assertEqual('EntityNotFound', resp.json['error']['type'])

        mock_call.assert_called_once_with(
            req.context,
            ('describe_stack_resource',
             {'stack_identity': stack_identity, 'resource_name': res_name,
              'with_attr': False}),
            version='1.2'
        )

    def test_metadata_show_nonexist_resource(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'metadata', True)
        res_name = 'wibble'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(res_identity._tenant_path() + '/metadata')

        error = heat_exc.ResourceNotFound(stack_name='a', resource_name='b')
        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     side_effect=tools.to_remote_error(error))

        resp = tools.request_with_middleware(
            fault.FaultWrapper,
            self.controller.metadata,
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id,
            resource_name=res_name)

        self.assertEqual(404, resp.json['code'])
        self.assertEqual('ResourceNotFound', resp.json['error']['type'])

        mock_call.assert_called_once_with(
            req.context,
            ('describe_stack_resource',
             {'stack_identity': stack_identity, 'resource_name': res_name,
              'with_attr': False}),
            version='1.2'
        )

    def test_metadata_show_err_denied_policy(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'metadata', False)
        res_name = 'wibble'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')
        res_identity = identifier.ResourceIdentifier(resource_name=res_name,
                                                     **stack_identity)

        req = self._get(res_identity._tenant_path() + '/metadata')

        resp = tools.request_with_middleware(
            fault.FaultWrapper,
            self.controller.metadata,
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id,
            resource_name=res_name)

        self.assertEqual(403, resp.status_int)
        self.assertIn('403 Forbidden', str(resp))

    def test_signal(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'signal', True)
        res_name = 'WikiDatabase'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '6')

        req = self._get(stack_identity._tenant_path())

        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=None)

        result = self.controller.signal(req, tenant_id=self.tenant,
                                        stack_name=stack_identity.stack_name,
                                        stack_id=stack_identity.stack_id,
                                        resource_name=res_name,
                                        body="Signal content")

        self.assertIsNone(result)

        mock_call.assert_called_once_with(
            req.context,
            ('resource_signal', {'stack_identity': stack_identity,
                                 'resource_name': res_name,
                                 'details': 'Signal content',
                                 'sync_call': False}),
            version='1.3'
        )

    def test_mark_unhealthy_valid_request(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'mark_unhealthy', True)
        res_name = 'WebServer'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')

        req = self._get(stack_identity._tenant_path())

        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=None)
        body = {'mark_unhealthy': True,
                rpc_api.RES_STATUS_DATA: 'Anything'}
        params = {'stack_identity': stack_identity,
                  'resource_name': res_name}
        params.update(body)

        result = self.controller.mark_unhealthy(
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id,
            resource_name=res_name,
            body=body)

        self.assertIsNone(result)

        mock_call.assert_called_once_with(
            req.context,
            ('resource_mark_unhealthy', params),
            version='1.26'
        )

    def test_mark_unhealthy_without_reason(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'mark_unhealthy', True)
        res_name = 'WebServer'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')

        req = self._get(stack_identity._tenant_path())

        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=None)
        body = {'mark_unhealthy': True, rpc_api.RES_STATUS_DATA: ''}
        params = {'stack_identity': stack_identity,
                  'resource_name': res_name}
        params.update(body)

        del body[rpc_api.RES_STATUS_DATA]

        result = self.controller.mark_unhealthy(
            req, tenant_id=self.tenant,
            stack_name=stack_identity.stack_name,
            stack_id=stack_identity.stack_id,
            resource_name=res_name,
            body=body)

        self.assertIsNone(result)

        mock_call.assert_called_once_with(
            req.context,
            ('resource_mark_unhealthy', params),
            version='1.26'
        )

    def test_mark_unhealthy_with_invalid_keys(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'mark_unhealthy', True)
        res_name = 'WebServer'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')

        req = self._get(stack_identity._tenant_path())

        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=None)
        body = {'mark_unhealthy': True,
                rpc_api.RES_STATUS_DATA: 'Any',
                'invalid_key1': 1, 'invalid_key2': 2}
        expected = "Invalid keys in resource mark unhealthy"
        actual = self.assertRaises(webob.exc.HTTPBadRequest,
                                   self.controller.mark_unhealthy, req,
                                   tenant_id=self.tenant,
                                   stack_name=stack_identity.stack_name,
                                   stack_id=stack_identity.stack_id,
                                   resource_name=res_name,
                                   body=body)

        self.assertIn(expected, str(actual))
        self.assertIn('invalid_key1', str(actual))
        self.assertIn('invalid_key2', str(actual))
        mock_call.assert_not_called()

    def test_mark_unhealthy_with_invalid_value(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'mark_unhealthy', True)
        res_name = 'WebServer'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')

        req = self._get(stack_identity._tenant_path())

        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=None)
        body = {'mark_unhealthy': 'XYZ',
                rpc_api.RES_STATUS_DATA: 'Any'}

        expected = ('Unrecognized value "XYZ" for "mark_unhealthy", '
                    'acceptable values are: true, false')

        actual = self.assertRaises(webob.exc.HTTPBadRequest,
                                   self.controller.mark_unhealthy, req,
                                   tenant_id=self.tenant,
                                   stack_name=stack_identity.stack_name,
                                   stack_id=stack_identity.stack_id,
                                   resource_name=res_name,
                                   body=body)

        self.assertIn(expected, str(actual))
        mock_call.assert_not_called()

    def test_mark_unhealthy_without_mark_unhealthy_key(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'mark_unhealthy', True)
        res_name = 'WebServer'
        stack_identity = identifier.HeatIdentifier(self.tenant,
                                                   'wordpress', '1')

        req = self._get(stack_identity._tenant_path())

        mock_call = self.patchobject(rpc_client.EngineClient, 'call',
                                     return_value=None)
        body = {rpc_api.RES_STATUS_DATA: 'Any'}

        expected = ("Missing mandatory (%s) key from "
                    "mark unhealthy request" % 'mark_unhealthy')

        actual = self.assertRaises(webob.exc.HTTPBadRequest,
                                   self.controller.mark_unhealthy, req,
                                   tenant_id=self.tenant,
                                   stack_name=stack_identity.stack_name,
                                   stack_id=stack_identity.stack_id,
                                   resource_name=res_name,
                                   body=body)

        self.assertIn(expected, str(actual))
        mock_call.assert_not_called()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import weakref

from heat.objects import resource as resource_object


class ParentResourceProxy(object):
    """Proxy for the TemplateResource that owns a provider stack.

    This is an interface through which the Fn::ResourceFacade/resource_facade
    intrinsic functions in a stack can access data about the TemplateResource
    in the parent stack for which it was created.

    This API can be considered stable by third-party Function plugins, and no
    part of it should be changed or removed without an appropriate deprecation
    process.
    """

    def __new__(cls, context, parent_resource_name, parent_stack_id):
        if parent_resource_name is None:
            return None
        return super(ParentResourceProxy, cls).__new__(cls)

    def __init__(self, context, parent_resource_name, parent_stack_id):
        self._context = context
        self.name = parent_resource_name
        self._stack_id = parent_stack_id
        self._stack_ref = None
        self._parent_stack = None

    def _stack(self):
        if self._stack_ref is not None:
            stk = self._stack_ref()
            if stk is not None:
                return stk

        assert self._stack_id is not None, "Must provide parent stack or ID"

        from heat.engine import stack
        self._parent_stack = stack.Stack.load(self._context,
                                              stack_id=self._stack_id)
        self._stack_ref = weakref.ref(self._parent_stack)
        return self._parent_stack

    def metadata_get(self):
        """Return the resource metadata."""
        # If we're using an existing stack that was passed in, assume that its
        # resources are already in memory. If they haven't been stored to the
        # DB yet, this avoids an unnecessary attempt to read from it.
        if self._parent_stack is None:
            refd_stk = self._stack_ref and self._stack_ref()
            if refd_stk is not None:
                return refd_stk[self.name].metadata_get()

        assert self._stack_id is not None, "Must provide parent stack or ID"

        # Try to read just this resource from the DB
        rs = resource_object.Resource.get_by_name_and_stack(self._context,
                                                            self.name,
                                                            self._stack_id)
        if rs is not None:
            return rs.rsrc_metadata

        # Resource not stored, just return the data from the template
        return self.t.metadata()

    @property
    def t(self):
        """The resource definition."""
        stk = self._stack()
        return stk.t.resource_definitions(stk)[self.name]


def use_parent_stack(parent_proxy, stack):
    parent_proxy._stack_ref = weakref.ref(stack)
    parent_proxy._parent_stack = None



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
from heat.engine import translation


class Port(resource.Resource):
    """A resource that creates a ironic port.

    Node UUID and physical hardware address for the Port (MAC address in
    most cases) are needed (all Ports must be associated to a Node when
    created).
    """

    support_status = support.SupportStatus(version='13.0.0')

    default_client_name = 'ironic'

    entity = 'port'

    PROPERTIES = (
        NODE, ADDRESS, PORTGROUP, LOCAL_LINK_CONNECTION, PXE_ENABLED,
        PHYSICAL_NETWORK, EXTRA, IS_SMARTNIC,
    ) = (
        'node', 'address', 'portgroup', 'local_link_connection', 'pxe_enabled',
        'physical_network', 'extra', 'is_smartnic',
    )
    PROPERTIES_MIN_SUPPORT_VERSION = (
        (PXE_ENABLED, '1.19'),
        (LOCAL_LINK_CONNECTION, '1.19'),
        (PORTGROUP, '1.24'),
        (PHYSICAL_NETWORK, '1.34'),
        (IS_SMARTNIC, '1.53')
    )

    ATTRIBUTES = (
        ADDRESS_ATTR, NODE_UUID_ATTR, PORTGROUP_UUID_ATTR,
        LOCAL_LINK_CONNECTION_ATTR, PXE_ENABLED_ATTR, PHYSICAL_NETWORK_ATTR,
        INTERNAL_INFO_ATTR, EXTRA_ATTR, IS_SMARTNIC_ATTR,
    ) = (
        'address', 'node_uuid', 'portgroup_uuid',
        'local_link_connection', 'pxe_enabled', 'physical_network',
        'internal_info', 'extra', 'is_smartnic',
    )
    attributes_schema = {
        ADDRESS_ATTR: attributes.Schema(
            _('Physical hardware address of this network Port, typically the '
              'hardware MAC address.'),
            type=attributes.Schema.STRING
        ),
        NODE_UUID_ATTR: attributes.Schema(
            _('UUID of the Node this resource belongs to.'),
            type=attributes.Schema.STRING
        ),
        PORTGROUP_UUID_ATTR: attributes.Schema(
            _('UUID of the Portgroup this resource belongs to.'),
            type=attributes.Schema.STRING
        ),
        LOCAL_LINK_CONNECTION_ATTR: attributes.Schema(
            _('The Port binding profile. If specified, must contain switch_id '
              '(only a MAC address or an OpenFlow based datapath_id of the '
              'switch are accepted in this field) and port_id (identifier of '
              'the physical port on the switch to which node\'s port is '
              'connected to) fields. switch_info is an optional string field '
              'to be used to store any vendor-specific information.'),
            type=attributes.Schema.MAP
        ),
        PXE_ENABLED_ATTR: attributes.Schema(
            _('Indicates whether PXE is enabled or disabled on the Port.'),
            type=attributes.Schema.BOOLEAN
        ),
        PHYSICAL_NETWORK_ATTR: attributes.Schema(
            _('The name of the physical network to which a port is connected. '
              'May be empty.'),
            type=attributes.Schema.STRING
        ),
        INTERNAL_INFO_ATTR: attributes.Schema(
            _('Internal metadata set and stored by the Port. This field is '
              'read-only.'),
            type=attributes.Schema.MAP
        ),
        EXTRA_ATTR: attributes.Schema(
            _('A set of one or more arbitrary metadata key and value pairs.'),
            type=attributes.Schema.MAP
        ),
        IS_SMARTNIC_ATTR: attributes.Schema(
            _('Indicates whether the Port is a Smart NIC port.'),
            type=attributes.Schema.BOOLEAN
        )}

    properties_schema = {
        NODE: properties.Schema(
            properties.Schema.STRING,
            _('UUID or name of the Node this resource belongs to.'),
            constraints=[
                constraints.CustomConstraint('ironic.node')
            ],
            required=True,
            update_allowed=True
        ),
        ADDRESS: properties.Schema(
            properties.Schema.STRING,
            _('Physical hardware address of this network Port, typically the '
              'hardware MAC address.'),
            required=True,
            update_allowed=True
        ),
        PORTGROUP: properties.Schema(
            properties.Schema.STRING,
            _('UUID or name of the Portgroup this resource belongs to.'),
            constraints=[
                constraints.CustomConstraint('ironic.portgroup')
            ],
            update_allowed=True,
        ),
        LOCAL_LINK_CONNECTION: properties.Schema(
            properties.Schema.MAP,
            _('The Port binding profile. If specified, must contain switch_id '
              '(only a MAC address or an OpenFlow based datapath_id of the '
              'switch are accepted in this field) and port_id (identifier of '
              'the physical port on the switch to which node\'s port is '
              'connected to) fields. switch_info is an optional string field '
              'to be used to store any vendor-specific information.'),
            update_allowed=True,
        ),
        PXE_ENABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Indicates whether PXE is enabled or disabled on the Port.'),
            update_allowed=True,
        ),
        PHYSICAL_NETWORK: properties.Schema(
            properties.Schema.STRING,
            _('The name of the physical network to which a port is connected. '
              'May be empty.'),
            update_allowed=True,
        ),
        EXTRA: properties.Schema(
            properties.Schema.MAP,
            _('A set of one or more arbitrary metadata key and value pairs.'),
            update_allowed=True,
        ),
        IS_SMARTNIC: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Indicates whether the Port is a Smart NIC port.'),
            update_allowed=True,
        )
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.NODE],
                client_plugin=self.client_plugin('ironic'),
                finder='get_node'),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.PORTGROUP],
                client_plugin=self.client_plugin('ironic'),
                finder='get_portgroup'),
        ]

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        port = self.client().port.get(self.resource_id)
        return getattr(port, name, None)

    def _check_supported(self, properties):
        # TODO(ricolin) Implement version support in property schema.
        for k, v in self.PROPERTIES_MIN_SUPPORT_VERSION:
            if k in properties and properties[k] is not None and (
                not self.client_plugin().is_version_supported(v)
            ):
                raise exception.NotSupported(
                    feature="OS::Ironic::Port with %s property" % k)

    def handle_create(self):
        args = dict(self.properties.items())
        self._check_supported(args)
        args['node_uuid'] = args.pop(self.NODE)
        if self.PORTGROUP in args:
            args['portgroup_uuid'] = args.pop(self.PORTGROUP)
        port = self.client().port.create(**args)
        self.resource_id_set(port.uuid)
        return port.uuid

    def check_create_complete(self, id):
        try:
            self.client().port.get(id)
        except Exception as exc:
            self.client_plugin().ignore_not_found(exc)
            return False
        return True

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self._check_supported(prop_diff)
            if self.NODE in prop_diff:
                prop_diff['node_uuid'] = prop_diff.pop(self.NODE)
            if self.PORTGROUP in prop_diff:
                prop_diff['portgroup_uuid'] = prop_diff.pop(self.PORTGROUP)
            patch = [{'op': 'replace', 'path': '/' + k, 'value': v}
                     for k, v in prop_diff.items()]
            self.client().port.update(self.resource_id, patch)
            return self.resource_id, prop_diff

    def check_delete_complete(self, id):
        if not id:
            return True
        try:
            self.client().port.get(id)
        except Exception as exc:
            self.client_plugin().ignore_not_found(exc)
            return True
        return False


def resource_mapping():
    return {
        'OS::Ironic::Port': Port,
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import math

ADJUSTMENT_TYPES = (
    EXACT_CAPACITY, CHANGE_IN_CAPACITY, PERCENT_CHANGE_IN_CAPACITY) = (
    'exact_capacity', 'change_in_capacity', 'percent_change_in_capacity')

CFN_ADJUSTMENT_TYPES = (
    CFN_EXACT_CAPACITY, CFN_CHANGE_IN_CAPACITY,
    CFN_PERCENT_CHANGE_IN_CAPACITY) = ('ExactCapacity', 'ChangeInCapacity',
                                       'PercentChangeInCapacity')


def calculate_new_capacity(current, adjustment, adjustment_type,
                           min_adjustment_step, minimum, maximum):
    """Calculates new capacity from the given adjustments.

    Given the current capacity, calculates the new capacity which results
    from applying the given adjustment of the given adjustment-type.  The
    new capacity will be kept within the maximum and minimum bounds.
    """
    def _get_minimum_adjustment(adjustment, min_adjustment_step):
        if min_adjustment_step and min_adjustment_step > abs(adjustment):
            adjustment = (min_adjustment_step if adjustment > 0
                          else -min_adjustment_step)
        return adjustment

    if adjustment_type in (CHANGE_IN_CAPACITY, CFN_CHANGE_IN_CAPACITY):
        new_capacity = current + adjustment
    elif adjustment_type in (EXACT_CAPACITY, CFN_EXACT_CAPACITY):
        new_capacity = adjustment
    else:
        # PercentChangeInCapacity
        delta = current * adjustment / 100.0
        if math.fabs(delta) < 1.0:
            rounded = int(math.ceil(delta) if delta > 0.0
                          else math.floor(delta))
        else:
            rounded = int(math.floor(delta) if delta > 0.0
                          else math.ceil(delta))
        adjustment = _get_minimum_adjustment(rounded, min_adjustment_step)
        new_capacity = current + adjustment

    if new_capacity > maximum:
        return maximum

    if new_capacity < minimum:
        return minimum

    return new_capacity



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from heat.common import exception
from heat.engine.cfn import functions as cfn_funcs
from heat.engine import function
from heat.engine.hot import functions as hot_funcs
from heat.engine import parameters
from heat.engine import properties
from heat.engine import translation
from heat.tests import common


class TestTranslationRule(common.HeatTestCase):

    def setUp(self):
        super(TestTranslationRule, self).setUp()
        self.props = mock.Mock(spec=properties.Properties)

    def test_translation_rule(self):
        for r in translation.TranslationRule.RULE_KEYS:
            props = properties.Properties({}, {})
            rule = translation.TranslationRule(
                props,
                r,
                ['any'],
                ['value'] if r == 'Add' else None,
                'value_name' if r == 'Replace' else None,
                'client_plugin' if r == 'Resolve' else None,
                'finder' if r == 'Resolve' else None)
            self.assertEqual(rule.properties, props)
            self.assertEqual(rule.rule, r)
            if r == 'Add':
                self.assertEqual(['value'], rule.value)
            if r == 'Replace':
                self.assertEqual('value_name', rule.value_name)
            else:
                self.assertIsNone(rule.value_name)

    def test_cmp_rules(self):
        rules = [
            translation.TranslationRule(
                mock.Mock(spec=properties.Properties),
                translation.TranslationRule.DELETE,
                ['any']
            ),
            translation.TranslationRule(
                mock.Mock(spec=properties.Properties),
                translation.TranslationRule.ADD,
                ['any']
            ),
            translation.TranslationRule(
                mock.Mock(spec=properties.Properties),
                translation.TranslationRule.RESOLVE,
                ['any'],
                client_plugin=mock.ANY,
                finder=mock.ANY
            ),
            translation.TranslationRule(
                mock.Mock(spec=properties.Properties),
                translation.TranslationRule.REPLACE,
                ['any']
            )
        ]
        expected = [translation.TranslationRule.ADD,
                    translation.TranslationRule.REPLACE,
                    translation.TranslationRule.RESOLVE,
                    translation.TranslationRule.DELETE]
        result = [rule.rule for rule in sorted(rules)]
        self.assertEqual(expected, result)

    def test_invalid_translation_rule(self):
        props = properties.Properties({}, {})

        exc = self.assertRaises(ValueError,
                                translation.TranslationRule,
                                props,
                                'EatTheCookie',
                                mock.ANY,
                                mock.ANY)
        self.assertEqual('There is no rule EatTheCookie. List of allowed '
                         'rules is: Add, Replace, Delete, Resolve.',
                         str(exc))

        exc = self.assertRaises(ValueError,
                                translation.TranslationRule,
                                props,
                                translation.TranslationRule.ADD,
                                'networks.network',
                                'value')
        self.assertEqual('"translation_path" should be non-empty list '
                         'with path to translate.',
                         str(exc))

        exc = self.assertRaises(ValueError,
                                translation.TranslationRule,
                                props,
                                translation.TranslationRule.ADD,
                                [],
                                mock.ANY)
        self.assertEqual('"translation_path" should be non-empty list '
                         'with path to translate.',
                         str(exc))

        exc = self.assertRaises(ValueError,
                                translation.TranslationRule,
                                props,
                                translation.TranslationRule.ADD,
                                ['any'],
                                'value',
                                'value_name',
                                'some_path')
        self.assertEqual('"value_path", "value" and "value_name" are '
                         'mutually exclusive and cannot be specified '
                         'at the same time.', str(exc))

        exc = self.assertRaises(ValueError,
                                translation.TranslationRule,
                                props,
                                translation.TranslationRule.ADD,
                                ['any'],
                                'value')
        self.assertEqual('"value" must be list type when rule is Add.',
                         str(exc))

    def test_add_rule_exist(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.LIST,
                schema=properties.Schema(
                    properties.Schema.MAP,
                    schema={
                        'red': properties.Schema(
                            properties.Schema.STRING
                        )
                    }
                )
            ),
            'bar': properties.Schema(
                properties.Schema.STRING
            )}

        data = {
            'far': [
                {'red': 'blue'}
            ],
            'bar': 'dak'
        }
        props = properties.Properties(schema, copy.copy(data))

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.ADD,
            ['far'],
            [{'red': props.get('bar')}])

        tran = translation.Translation(props)
        tran.set_rules([rule])

        self.assertTrue(tran.has_translation('far'))
        result = tran.translate('far', data['far'])
        self.assertEqual([{'red': 'blue'}, {'red': 'dak'}], result)
        self.assertEqual([{'red': 'blue'}, {'red': 'dak'}],
                         tran.resolved_translations['far'])

    def test_add_rule_dont_exist(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.LIST,
                schema=properties.Schema(
                    properties.Schema.MAP,
                    schema={
                        'red': properties.Schema(
                            properties.Schema.STRING
                        )
                    }
                )
            ),
            'bar': properties.Schema(
                properties.Schema.STRING
            )}

        data = {
            'bar': 'dak'
        }
        props = properties.Properties(schema, copy.copy(data))

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.ADD,
            ['far'],
            [{'red': props.get('bar')}])

        tran = translation.Translation(props)
        tran.set_rules([rule])

        self.assertTrue(tran.has_translation('far'))
        result = tran.translate('far')
        self.assertEqual([{'red': 'dak'}], result)
        self.assertEqual([{'red': 'dak'}], tran.resolved_translations['far'])

    def test_add_rule_invalid(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.MAP,
                schema={
                    'red': properties.Schema(
                        properties.Schema.STRING
                    )
                }
            ),
            'bar': properties.Schema(
                properties.Schema.STRING
            )}

        data = {
            'far': 'tran',
            'bar': 'dak'
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.ADD,
            ['far'],
            [props.get('bar')])

        tran = translation.Translation(props)
        tran.set_rules([rule])

        self.assertTrue(tran.has_translation('far'))
        ex = self.assertRaises(ValueError, tran.translate, 'far', 'tran')
        self.assertEqual('Incorrect translation rule using - cannot '
                         'resolve Add rule for non-list translation '
                         'value "far".', str(ex))

    def test_replace_rule_map_exist(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.MAP,
                schema={
                    'red': properties.Schema(
                        properties.Schema.STRING
                    )
                }
            ),
            'bar': properties.Schema(
                properties.Schema.STRING
            )}

        data = {
            'far': {'red': 'tran'},
            'bar': 'dak'
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['far', 'red'],
            props.get('bar'))

        tran = translation.Translation(props)
        tran.set_rules([rule])

        self.assertTrue(tran.has_translation('far.red'))
        result = tran.translate('far.red', data['far']['red'])
        self.assertEqual('dak', result)
        self.assertEqual('dak', tran.resolved_translations['far.red'])

    def test_replace_rule_map_dont_exist(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.MAP,
                schema={
                    'red': properties.Schema(
                        properties.Schema.STRING
                    )
                }
            ),
            'bar': properties.Schema(
                properties.Schema.STRING
            )}

        data = {
            'bar': 'dak'
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['far', 'red'],
            props.get('bar'))

        tran = translation.Translation(props)
        tran.set_rules([rule])

        self.assertTrue(tran.has_translation('far.red'))
        result = tran.translate('far.red')
        self.assertEqual('dak', result)
        self.assertEqual('dak', tran.resolved_translations['far.red'])

    def test_replace_rule_list_different(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.LIST,
                schema=properties.Schema(
                    properties.Schema.MAP,
                    schema={
                        'red': properties.Schema(
                            properties.Schema.STRING
                        )
                    }
                )
            ),
            'bar': properties.Schema(
                properties.Schema.STRING
            )}

        data = {
            'far': [{'red': 'blue'},
                    {'red': 'roses'}],
            'bar': 'dak'
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['far', 'red'],
            props.get('bar'))

        tran = translation.Translation(props)
        tran.set_rules([rule])

        self.assertTrue(tran.has_translation('far.red'))
        result = tran.translate('far.0.red', data['far'][0]['red'])
        self.assertEqual('dak', result)
        self.assertEqual('dak', tran.resolved_translations['far.0.red'])

    def test_replace_rule_list_same(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.LIST,
                schema=properties.Schema(
                    properties.Schema.MAP,
                    schema={
                        'red': properties.Schema(
                            properties.Schema.STRING
                        ),
                        'blue': properties.Schema(
                            properties.Schema.STRING
                        )
                    }
                )
            )}

        data = {
            'far': [{'blue': 'white'},
                    {'red': 'roses'}]
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['far', 'red'],
            None,
            'blue')

        tran = translation.Translation(props)
        tran.set_rules([rule])

        self.assertTrue(tran.has_translation('far.0.red'))
        result = tran.translate('far.0.red', data['far'][0].get('red'),
                                data['far'][0])
        self.assertEqual('white', result)
        self.assertEqual('white', tran.resolved_translations['far.0.red'])
        self.assertIsNone(tran.resolved_translations['far.0.blue'])
        self.assertTrue(tran.has_translation('far.1.red'))
        result = tran.translate('far.1.red', data['far'][1]['red'],
                                data['far'][1])
        self.assertEqual('roses', result)
        self.assertEqual('roses', tran.resolved_translations['far.1.red'])
        self.assertIsNone(tran.resolved_translations['far.1.blue'])

    def test_replace_rule_str(self):
        schema = {
            'far': properties.Schema(properties.Schema.STRING),
            'bar': properties.Schema(properties.Schema.STRING)
        }

        data = {'far': 'one', 'bar': 'two'}

        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['bar'],
            props.get('far'))

        tran = translation.Translation(props)
        tran.set_rules([rule])

        self.assertTrue(tran.has_translation('bar'))
        result = tran.translate('bar', data['bar'])
        self.assertEqual('one', result)
        self.assertEqual('one', tran.resolved_translations['bar'])

    def test_replace_rule_str_value_path_error(self):
        schema = {
            'far': properties.Schema(properties.Schema.STRING),
            'bar': properties.Schema(properties.Schema.STRING)
        }

        data = {'far': 'one', 'bar': 'two'}

        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['bar'],
            value_path=['far'])

        tran = translation.Translation(props)
        tran.set_rules([rule])

        self.assertTrue(tran.has_translation('bar'))
        ex = self.assertRaises(exception.StackValidationFailed,
                               tran.translate, 'bar', data['bar'])
        self.assertEqual('Cannot define the following properties at '
                         'the same time: bar, far', str(ex))

    def test_replace_rule_str_value_path(self):
        schema = {
            'far': properties.Schema(properties.Schema.STRING),
            'bar': properties.Schema(properties.Schema.STRING)
        }

        props = properties.Properties(schema, {'far': 'one'})

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['bar'],
            value_path=['far'])

        props = properties.Properties(schema, {'far': 'one'})
        tran = translation.Translation(props)
        tran.set_rules([rule])

        self.assertTrue(tran.has_translation('bar'))
        result = tran.translate('bar')
        self.assertEqual('one', result)
        self.assertEqual('one', tran.resolved_translations['bar'])
        self.assertIsNone(tran.resolved_translations['far'])

    def test_replace_rule_str_invalid(self):
        schema = {
            'far': properties.Schema(properties.Schema.STRING),
            'bar': properties.Schema(properties.Schema.INTEGER)
        }

        data = {'far': 'one', 'bar': 2}

        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['bar'],
            props.get('far'))
        props.update_translation([rule])
        exc = self.assertRaises(exception.StackValidationFailed,
                                props.validate)
        self.assertEqual("Property error: bar: Value 'one' is not an integer",
                         str(exc))

    def test_delete_rule_list(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.LIST,
                schema=properties.Schema(
                    properties.Schema.MAP,
                    schema={
                        'red': properties.Schema(
                            properties.Schema.STRING
                        ),
                        'check': properties.Schema(
                            properties.Schema.STRING
                        )
                    }
                )
            )}

        data = {
            'far': [{'red': 'blue'},
                    {'red': 'roses'}],
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.DELETE,
            ['far', 'red'])

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far.red'))
        self.assertIsNone(tran.translate('far.red'))
        self.assertIsNone(tran.resolved_translations['far.red'])

    def test_delete_rule_other(self):
        schema = {
            'far': properties.Schema(properties.Schema.STRING)
        }

        data = {'far': 'one'}

        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.DELETE,
            ['far'])

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far'))
        self.assertIsNone(tran.translate('far'))
        self.assertIsNone(tran.resolved_translations['far'])

    def _test_resolve_rule(self, is_list=False,
                           check_error=False):
        class FakeClientPlugin(object):
            def find_name_id(self, entity=None,
                             src_value='far'):
                if check_error:
                    raise exception.NotFound()
                if entity == 'rose':
                    return 'pink'
                return 'yellow'

        if is_list:
            schema = {
                'far': properties.Schema(
                    properties.Schema.LIST,
                    schema=properties.Schema(
                        properties.Schema.MAP,
                        schema={
                            'red': properties.Schema(
                                properties.Schema.STRING
                            )
                        }
                    )
                )}
        else:
            schema = {
                'far': properties.Schema(properties.Schema.STRING)
            }
        return FakeClientPlugin(), schema

    def test_resolve_rule_list_populated(self):
        client_plugin, schema = self._test_resolve_rule(is_list=True)
        data = {
            'far': [{'red': 'blue'},
                    {'red': 'roses'}],
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['far', 'red'],
            client_plugin=client_plugin,
            finder='find_name_id'
            )

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far.red'))
        result = tran.translate('far.0.red', data['far'][0]['red'])
        self.assertEqual('yellow', result)
        self.assertEqual('yellow', tran.resolved_translations['far.0.red'])

    def test_resolve_rule_nested_list_populated(self):
        client_plugin, schema = self._test_resolve_rule_nested_list()
        data = {
            'instances': [{'networks': [{'port': 'port1', 'net': 'net1'}]}]
        }
        props = properties.Properties(schema, data)
        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['instances', 'networks', 'port'],
            client_plugin=client_plugin,
            finder='find_name_id',
            entity='port'
        )
        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('instances.networks.port'))
        result = tran.translate('instances.0.networks.0.port',
                                data['instances'][0]['networks'][0]['port'])
        self.assertEqual('port1_id', result)
        self.assertEqual('port1_id', tran.resolved_translations[
            'instances.0.networks.0.port'])

    def _test_resolve_rule_nested_list(self):
        class FakeClientPlugin(object):
            def find_name_id(self, entity=None, value=None):
                if entity == 'net':
                    return 'net1_id'
                if entity == 'port':
                    return 'port1_id'

        schema = {
            'instances': properties.Schema(
                properties.Schema.LIST,
                schema=properties.Schema(
                    properties.Schema.MAP,
                    schema={
                        'networks': properties.Schema(
                            properties.Schema.LIST,
                            schema=properties.Schema(
                                properties.Schema.MAP,
                                schema={
                                    'port': properties.Schema(
                                        properties.Schema.STRING,
                                    ),
                                    'net': properties.Schema(
                                        properties.Schema.STRING,
                                    ),
                                }
                            )
                        )
                    }
                )
            )}

        return FakeClientPlugin(), schema

    def test_resolve_rule_list_with_function(self):
        client_plugin, schema = self._test_resolve_rule(is_list=True)
        join_func = cfn_funcs.Join(None,
                                   'Fn::Join', ['.', ['bar', 'baz']])
        data = {
            'far': [{'red': 'blue'},
                    {'red': join_func}],
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['far', 'red'],
            client_plugin=client_plugin,
            finder='find_name_id'
            )

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far.red'))
        result = tran.translate('far.0.red', data['far'][0]['red'])
        self.assertEqual('yellow', result)
        self.assertEqual('yellow', tran.resolved_translations['far.0.red'])

    def test_resolve_rule_list_with_ref(self):
        client_plugin, schema = self._test_resolve_rule(is_list=True)

        class rsrc(object):
            action = INIT = "INIT"

            def FnGetRefId(self):
                return 'resource_id'

        class DummyStack(dict):
            pass

        stack = DummyStack(another_res=rsrc())
        ref = hot_funcs.GetResource(stack, 'get_resource',
                                    'another_res')
        data = {
            'far': [{'red': ref}],
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['far', 'red'],
            client_plugin=client_plugin,
            finder='find_name_id'
            )

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far.red'))
        result = tran.translate('far.0.red', data['far'][0]['red'])
        self.assertEqual('yellow', result)
        self.assertEqual('yellow', tran.resolved_translations['far.0.red'])

    def test_resolve_rule_list_strings(self):
        client_plugin, schema = self._test_resolve_rule()
        data = {'far': ['one', 'rose']}
        schema = {'far': properties.Schema(
            properties.Schema.LIST,
            schema=properties.Schema(
                properties.Schema.STRING
            )
        )}
        props = properties.Properties(schema, data)
        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['far'],
            client_plugin=client_plugin,
            finder='find_name_id')

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far'))
        result = tran.translate('far', data['far'])
        self.assertEqual(['yellow', 'pink'], result)
        self.assertEqual(['yellow', 'pink'], tran.resolved_translations['far'])

    def test_resolve_rule_ignore_error(self):
        client_plugin, schema = self._test_resolve_rule(check_error=True)
        data = {'far': 'one'}
        props = properties.Properties(schema, data)
        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['far'],
            client_plugin=client_plugin,
            finder='find_name_id')

        tran = translation.Translation(props)
        tran.set_rules([rule], ignore_resolve_error=True)
        self.assertTrue(tran.has_translation('far'))
        result = tran.translate('far', data['far'])
        self.assertEqual('one', result)
        self.assertEqual('one', tran.resolved_translations['far'])

    def test_resolve_rule_other(self):
        client_plugin, schema = self._test_resolve_rule()
        data = {'far': 'one'}
        props = properties.Properties(schema, data)
        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['far'],
            client_plugin=client_plugin,
            finder='find_name_id')

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far'))
        result = tran.translate('far', data['far'])
        self.assertEqual('yellow', result)
        self.assertEqual('yellow', tran.resolved_translations['far'])

    def test_resolve_rule_other_with_ref(self):
        client_plugin, schema = self._test_resolve_rule()

        class rsrc(object):
            action = INIT = "INIT"

            def FnGetRefId(self):
                return 'resource_id'

        class DummyStack(dict):
            pass

        stack = DummyStack(another_res=rsrc())
        ref = hot_funcs.GetResource(stack, 'get_resource',
                                    'another_res')
        data = {'far': ref}
        props = properties.Properties(schema, data)
        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['far'],
            client_plugin=client_plugin,
            finder='find_name_id')

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far'))
        result = tran.translate('far', data['far'])
        self.assertEqual('yellow', result)

    def test_resolve_rule_other_with_function(self):
        client_plugin, schema = self._test_resolve_rule()
        join_func = cfn_funcs.Join(None,
                                   'Fn::Join', ['.', ['bar', 'baz']])
        data = {'far': join_func}
        props = properties.Properties(schema, data)
        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['far'],
            client_plugin=client_plugin,
            finder='find_name_id')

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far'))
        result = tran.translate('far', data['far'])
        self.assertEqual('yellow', result)
        self.assertEqual('yellow', tran.resolved_translations['far'])

    def test_resolve_rule_other_with_get_attr(self):
        client_plugin, schema = self._test_resolve_rule()

        class DummyStack(dict):
            pass

        class rsrc(object):
            pass

        stack = DummyStack(another_res=rsrc())
        attr_func = cfn_funcs.GetAtt(stack, 'Fn::GetAtt',
                                     ['another_res', 'name'])
        data = {'far': attr_func}
        props = properties.Properties(schema, data)
        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['far'],
            client_plugin=client_plugin,
            finder='find_name_id')

        tran = translation.Translation(props)
        tran.set_rules([rule], client_resolve=False)
        self.assertFalse(tran.store_translated_values)
        self.assertFalse(tran.has_translation('far'))
        result = tran.translate('far', 'no_check', data['far'])
        self.assertEqual('no_check', result)
        self.assertIsNone(tran.resolved_translations.get('far'))

    def test_resolve_rule_other_with_entity(self):
        client_plugin, schema = self._test_resolve_rule()
        data = {'far': 'one'}
        props = properties.Properties(schema, data)
        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.RESOLVE,
            ['far'],
            client_plugin=client_plugin,
            finder='find_name_id',
            entity='rose')

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far'))
        result = tran.translate('far', data['far'])
        self.assertEqual('pink', result)
        self.assertEqual('pink', tran.resolved_translations['far'])

    def test_property_json_param_correct_translation(self):
        """Test case when property with sub-schema takes json param."""
        schema = {
            'far': properties.Schema(properties.Schema.MAP,
                                     schema={
                                         'bar': properties.Schema(
                                             properties.Schema.STRING,
                                         ),
                                         'dar': properties.Schema(
                                             properties.Schema.STRING
                                         )
                                     })
        }

        class DummyStack(dict):
            @property
            def parameters(self):
                return mock.Mock()

        param = hot_funcs.GetParam(DummyStack(json_far='json_far'),
                                   'get_param',
                                   'json_far')
        param.parameters = {
            'json_far': parameters.JsonParam(
                'json_far',
                {'Type': 'Json'},
                '{"dar": "rad"}').value()}
        data = {'far': param}

        props = properties.Properties(schema, data, resolver=function.resolve)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['far', 'bar'],
            value_path=['far', 'dar'])

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far.bar'))
        prop_data = props['far']
        result = tran.translate('far.bar', prop_data['bar'], prop_data)
        self.assertEqual('rad', result)
        self.assertEqual('rad', tran.resolved_translations['far.bar'])

    def test_property_json_param_to_list_correct_translation(self):
        """Test case when list property with sub-schema takes json param."""
        schema = {
            'far': properties.Schema(properties.Schema.LIST,
                                     schema=properties.Schema(
                                         properties.Schema.MAP,
                                         schema={
                                             'bar': properties.Schema(
                                                 properties.Schema.STRING,
                                             ),
                                             'dar': properties.Schema(
                                                 properties.Schema.STRING
                                             )
                                         }
                                     ))
        }

        class DummyStack(dict):
            @property
            def parameters(self):
                return mock.Mock()

        param = hot_funcs.GetParam(DummyStack(json_far='json_far'),
                                   'get_param',
                                   'json_far')
        param.parameters = {
            'json_far': parameters.JsonParam(
                'json_far',
                {'Type': 'Json'},
                '{"dar": "rad"}').value()}
        data = {'far': [param]}

        props = properties.Properties(schema, data, resolver=function.resolve)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['far', 'bar'],
            value_name='dar')

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far.0.bar'))
        prop_data = props['far']
        result = tran.translate('far.0.bar', prop_data[0]['bar'],
                                prop_data[0])
        self.assertEqual('rad', result)
        self.assertEqual('rad', tran.resolved_translations['far.0.bar'])

    def test_property_commadelimitedlist_param_correct_translation(self):
        """Test when property with sub-schema takes comma_delimited_list."""
        schema = {
            'far': properties.Schema(
                properties.Schema.LIST,
                schema=properties.Schema(
                    properties.Schema.STRING,
                )
            ),
            'boo': properties.Schema(
                properties.Schema.STRING
            )}

        class DummyStack(dict):
            @property
            def parameters(self):
                return mock.Mock()

        param = hot_funcs.GetParam(DummyStack(list_far='list_far'),
                                   'get_param',
                                   'list_far')
        param.parameters = {
            'list_far': parameters.CommaDelimitedListParam(
                'list_far',
                {'Type': 'CommaDelimitedList'},
                "white,roses").value()}
        data = {'far': param, 'boo': 'chrysanthemums'}

        props = properties.Properties(schema, data, resolver=function.resolve)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.ADD,
            ['far'],
            [props.get('boo')])

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far'))
        result = tran.translate('far', props['far'])
        self.assertEqual(['white', 'roses', 'chrysanthemums'], result)
        self.assertEqual(['white', 'roses', 'chrysanthemums'],
                         tran.resolved_translations['far'])

    def test_list_list_add_translation_rule(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.LIST,
                schema=properties.Schema(
                    properties.Schema.MAP,
                    schema={
                        'bar': properties.Schema(
                            properties.Schema.LIST,
                            schema=properties.Schema(properties.Schema.STRING)
                        ),
                        'car': properties.Schema(properties.Schema.STRING)
                    }
                )
            )
        }

        data = {'far': [{'bar': ['shar'], 'car': 'man'}, {'car': 'first'}]}

        props = properties.Properties(schema, data, resolver=function.resolve)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.ADD,
            ['far', 'bar'],
            value_name='car'
        )

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far.0.bar'))
        result = tran.translate('far.0.bar', props['far'][0]['bar'],
                                props['far'][0])
        self.assertEqual(['shar', 'man'], result)
        self.assertEqual(['shar', 'man'],
                         tran.resolved_translations['far.0.bar'])
        result = tran.translate('far.1.bar', prop_data=props['far'][1])
        self.assertEqual(['first'], result)
        self.assertEqual(['first'], tran.resolved_translations['far.1.bar'])

    def test_replace_rule_map_with_custom_value_path(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.MAP,
                schema={
                    'red': properties.Schema(
                        properties.Schema.STRING
                    )
                }
            ),
            'bar': properties.Schema(
                properties.Schema.MAP
            )}

        data = {
            'far': {},
            'bar': {'red': 'dak'}
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['far', 'red'],
            value_path=['bar'],
            custom_value_path=['red']
        )

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far.red'))
        result = tran.translate('far.red')
        self.assertEqual('dak', result)
        self.assertEqual('dak', tran.resolved_translations['far.red'])

    def test_replace_rule_list_with_custom_value_path(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.LIST,
                schema=properties.Schema(
                    properties.Schema.MAP,
                    schema={
                        'red': properties.Schema(
                            properties.Schema.STRING
                        ),
                        'blue': properties.Schema(
                            properties.Schema.MAP
                        )
                    }
                )
            )}

        data = {
            'far': [{'blue': {'black': {'white': 'daisy'}}},
                    {'red': 'roses'}]
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.REPLACE,
            ['far', 'red'],
            value_name='blue',
            custom_value_path=['black', 'white']
        )

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far.0.red'))
        result = tran.translate('far.0.red', prop_data=data['far'][0])
        self.assertEqual('daisy', result)
        self.assertEqual('daisy', tran.resolved_translations['far.0.red'])

    def test_add_rule_list_with_custom_value_path(self):
        schema = {
            'far': properties.Schema(
                properties.Schema.LIST,
                schema=properties.Schema(
                    properties.Schema.MAP,
                    schema={
                        'red': properties.Schema(
                            properties.Schema.LIST,
                            schema=properties.Schema(properties.Schema.STRING)
                        ),
                        'blue': properties.Schema(
                            properties.Schema.MAP
                        )
                    }
                )
            )}

        data = {
            'far': [{'blue': {'black': {'white': 'daisy', 'check': ['one']}}},
                    {'red': ['roses']}]
        }
        props = properties.Properties(schema, data)

        rule = translation.TranslationRule(
            props,
            translation.TranslationRule.ADD,
            ['far', 'red'],
            value_name='blue',
            custom_value_path=['black', 'check']
        )

        tran = translation.Translation(props)
        tran.set_rules([rule])
        self.assertTrue(tran.has_translation('far.0.red'))
        result = tran.translate('far.0.red', data['far'][0].get('red'),
                                data['far'][0])
        self.assertEqual(['one'], result)
        self.assertEqual(['one'], tran.resolved_translations['far.0.red'])
        self.assertEqual(['roses'], tran.translate('far.1.red',
                                                   data['far'][1]['red'],
                                                   data['far'][1]))

    def test_set_rules_none(self):
        tran = translation.Translation()
        self.assertEqual({}, tran._rules)

    def test_set_no_resolve_rules(self):
        rules = [
            translation.TranslationRule(
                self.props,
                translation.TranslationRule.RESOLVE,
                ['a'],
                client_plugin=mock.ANY,
                finder='finder'
            )
        ]

        tran = translation.Translation()
        tran.set_rules(rules, client_resolve=False)
        self.assertEqual({}, tran._rules)

    def test_translate_add(self):
        rules = [
            translation.TranslationRule(
                self.props,
                translation.TranslationRule.ADD,
                ['a', 'b'],
                value=['check']
            )
        ]

        tran = translation.Translation()
        tran.set_rules(rules)

        result = tran.translate('a.b', ['test'])
        self.assertEqual(['test', 'check'], result)
        self.assertEqual(['test', 'check'], tran.resolved_translations['a.b'])

        # Test without storing
        tran.resolved_translations = {}
        tran.store_translated_values = False
        result = tran.translate('a.b', ['test'])
        self.assertEqual(['test', 'check'], result)
        self.assertEqual({}, tran.resolved_translations)
        tran.store_translated_values = True

        # Test no prop_value
        self.assertEqual(['check'], tran.translate('a.b', None))

        # Check digits in path skipped for rule
        self.assertEqual(['test', 'check'], tran.translate('a.0.b', ['test']))

    def test_translate_delete(self):
        rules = [
            translation.TranslationRule(
                self.props,
                translation.TranslationRule.DELETE,
                ['a']
            )
        ]

        tran = translation.Translation()
        tran.set_rules(rules)

        self.assertIsNone(tran.translate('a'))
        self.assertIsNone(tran.resolved_translations['a'])

        # Test without storing
        tran.resolved_translations = {}
        tran.store_translated_values = False
        self.assertIsNone(tran.translate('a'))
        self.assertEqual({}, tran.resolved_translations)
        tran.store_translated_values = True



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat_integrationtests.functional import functional_base


class CancelUpdateTest(functional_base.FunctionalTestsBase):

    template = '''
heat_template_version: '2013-05-23'
parameters:
 InstanceType:
   type: string
 ImageId:
   type: string
 network:
   type: string
resources:
 port:
   type: OS::Neutron::Port
   properties:
     network: {get_param: network}
 Server:
   type: OS::Nova::Server
   properties:
     flavor_update_policy: REPLACE
     image: {get_param: ImageId}
     flavor: {get_param: InstanceType}
     networks:
       - port: {get_resource: port}
'''

    def setUp(self):
        super(CancelUpdateTest, self).setUp()
        if not self.conf.minimal_image_ref:
            raise self.skipException("No minimal image configured to test")
        if not self.conf.minimal_instance_type:
            raise self.skipException("No minimal flavor configured to test.")

    def test_cancel_update_server_with_port(self):
        parameters = {'InstanceType': self.conf.minimal_instance_type,
                      'ImageId': self.conf.minimal_image_ref,
                      'network': self.conf.fixed_network_name}

        stack_identifier = self.stack_create(template=self.template,
                                             parameters=parameters)
        parameters['InstanceType'] = self.conf.instance_type
        self.update_stack(stack_identifier, self.template,
                          parameters=parameters,
                          expected_status='UPDATE_IN_PROGRESS')

        # Ensure we start updating the server before rolling back
        self._wait_for_resource_status(
            stack_identifier, 'Server', 'CREATE_IN_PROGRESS')

        self.cancel_update_stack(stack_identifier)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutronclient.neutron import v2_0 as neutronV20
from osc_lib import exceptions

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os.octavia import OctaviaClientPlugin
from heat.engine.resources.openstack.octavia import loadbalancer
from heat.tests import common
from heat.tests.openstack.octavia import inline_templates
from heat.tests import utils


class LoadBalancerTest(common.HeatTestCase):

    def test_resource_mapping(self):
        mapping = loadbalancer.resource_mapping()
        self.assertEqual(loadbalancer.LoadBalancer,
                         mapping['OS::Octavia::LoadBalancer'])

    def _create_stack(self, tmpl=inline_templates.LB_TEMPLATE):
        self.t = template_format.parse(tmpl)
        self.stack = utils.parse_stack(self.t)
        self.lb = self.stack['lb']
        self.octavia_client = mock.MagicMock()
        self.lb.client = mock.MagicMock()
        self.lb.client.return_value = self.octavia_client

        self.patchobject(neutronV20, 'find_resourceid_by_name_or_id',
                         return_value='123')
        self.patchobject(OctaviaClientPlugin, 'get_flavor',
                         return_value='f123')

        self.lb.client_plugin().client = mock.MagicMock(
            return_value=self.octavia_client)
        self.lb.translate_properties(self.lb.properties)
        self.lb.resource_id_set('1234')

    def test_create(self):
        self._create_stack()
        expected = {
            'loadbalancer': {
                'name': 'my_lb',
                'description': 'my loadbalancer',
                'vip_address': '10.0.0.4',
                'vip_subnet_id': '123',
                'provider': 'octavia',
                'project_id': '1234',
                'admin_state_up': True,
                'flavor_id': 'f123',
                'availability_zone': 'nova',
            }
        }

        self.lb.handle_create()

        self.octavia_client.load_balancer_create.assert_called_with(
            json=expected)

    def test_check_create_complete(self):
        self._create_stack()
        self.octavia_client.load_balancer_show.side_effect = [
            {'provisioning_status': 'ACTIVE'},
            {'provisioning_status': 'PENDING_CREATE'},
            {'provisioning_status': 'ERROR'},
        ]

        self.assertTrue(self.lb.check_create_complete(None))
        self.assertFalse(self.lb.check_create_complete(None))
        self.assertRaises(exception.ResourceInError,
                          self.lb.check_create_complete, None)

    def test_show_resource(self):
        self._create_stack()
        self.octavia_client.load_balancer_show.return_value = {'id': '1234'}
        self.assertEqual({'id': '1234'}, self.lb._show_resource())

        self.octavia_client.load_balancer_show.assert_called_with('1234')

    def test_update(self):
        self._create_stack()
        prop_diff = {
            'name': 'lb',
            'description': 'a loadbalancer',
            'admin_state_up': False,
        }

        prop_diff = self.lb.handle_update(None, None, prop_diff)

        self.octavia_client.load_balancer_set.assert_called_once_with(
            '1234', json={'loadbalancer': prop_diff})

    def test_update_complete(self):
        self._create_stack()
        prop_diff = {
            'name': 'lb',
            'description': 'a loadbalancer',
            'admin_state_up': False,
        }
        self.octavia_client.load_balancer_show.side_effect = [
            {'provisioning_status': 'ACTIVE'},
            {'provisioning_status': 'PENDING_UPDATE'},
        ]

        self.lb.handle_update(None, None, prop_diff)

        self.assertTrue(self.lb.check_update_complete(prop_diff))
        self.assertFalse(self.lb.check_update_complete(prop_diff))
        self.assertTrue(self.lb.check_update_complete({}))

    def test_delete(self):
        self._create_stack()
        self.octavia_client.load_balancer_show.side_effect = [
            {'provisioning_status': 'DELETE_PENDING'},
            {'provisioning_status': 'DELETE_PENDING'},
            {'provisioning_status': 'DELETED'},
        ]

        self.octavia_client.load_balancer_delete.side_effect = [
            exceptions.Conflict(409),
            None
        ]

        self.lb.handle_delete()

        self.assertFalse(self.lb.check_delete_complete(None))
        self.assertFalse(self.lb._delete_called)
        self.assertFalse(self.lb.check_delete_complete(None))
        self.assertTrue(self.lb._delete_called)
        self.assertTrue(self.lb.check_delete_complete(None))
        self.octavia_client.load_balancer_delete.assert_called_with('1234')
        self.assertEqual(
            2, self.octavia_client.load_balancer_delete.call_count)

    def test_delete_error(self):
        self._create_stack()
        self.octavia_client.load_balancer_show.side_effect = [
            {'provisioning_status': 'DELETE_PENDING'},
        ]

        self.octavia_client.load_balancer_delete.side_effect = [
            exceptions.Conflict(409),
            exceptions.NotFound(404)
        ]

        self.lb.handle_delete()

        self.assertFalse(self.lb.check_delete_complete(None))
        self.assertTrue(self.lb.check_delete_complete(None))
        self.octavia_client.load_balancer_delete.assert_called_with('1234')
        self.assertEqual(
            2, self.octavia_client.load_balancer_delete.call_count)

    def test_delete_failed(self):
        self._create_stack()
        self.octavia_client.load_balancer_delete.side_effect = (
            exceptions.Unauthorized(403))

        self.lb.handle_delete()
        self.assertRaises(exceptions.Unauthorized,
                          self.lb.check_delete_complete, None)
